import json
import logging
import threading
import uuid
from functools import partial
from typing import Any, Dict, Iterable, List, Optional, Tuple

from dateutil import parser as dateutil_parser
from pyiceberg.catalog import Catalog
from pyiceberg.exceptions import (
    NoSuchIcebergTableError,
    NoSuchNamespaceError,
    NoSuchPropertyException,
    NoSuchTableError,
    ServerError,
)
from pyiceberg.schema import Schema, SchemaVisitorPerPrimitiveType, visit
from pyiceberg.table import Table
from pyiceberg.typedef import Identifier, Properties
from pyiceberg.types import (
    BinaryType,
    BooleanType,
    DateType,
    DecimalType,
    DoubleType,
    FixedType,
    FloatType,
    IntegerType,
    ListType,
    LongType,
    MapType,
    NestedField,
    StringType,
    StructType,
    TimestampType,
    TimestamptzType,
    TimeType,
    UUIDType,
)

from datahub.emitter.mce_builder import (
    make_container_urn,
    make_data_platform_urn,
    make_dataplatform_instance_urn,
    make_dataset_urn_with_platform_instance,
    make_group_urn,
    make_user_urn,
)
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.emitter.mcp_builder import NamespaceKey
from datahub.ingestion.api.auto_work_units.auto_dataset_properties_aspect import (
    auto_patch_last_modified,
)
from datahub.ingestion.api.auto_work_units.auto_ensure_aspect_size import (
    EnsureAspectSizeProcessor,
)
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.decorators import (
    SourceCapability,
    SupportStatus,
    capability,
    config_class,
    platform_name,
    support_status,
)
from datahub.ingestion.api.source import MetadataWorkUnitProcessor, SourceReport
from datahub.ingestion.api.source_helpers import (
    AutoSystemMetadata,
    auto_fix_duplicate_schema_field_paths,
    auto_fix_empty_field_paths,
    auto_lowercase_urns,
    auto_materialize_referenced_tags_terms,
    auto_workunit_reporter,
)
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.extractor import schema_util
from datahub.ingestion.source.common.subtypes import (
    DatasetContainerSubTypes,
    DatasetSubTypes,
)
from datahub.ingestion.source.iceberg.iceberg_common import (
    IcebergSourceConfig,
    IcebergSourceReport,
)
from datahub.ingestion.source.iceberg.iceberg_profiler import IcebergProfiler
from datahub.ingestion.source.state.stale_entity_removal_handler import (
    StaleEntityRemovalHandler,
)
from datahub.ingestion.source.state.stateful_ingestion_base import (
    StatefulIngestionSourceBase,
)
from datahub.metadata.com.linkedin.pegasus2avro.common import Status, SubTypes
from datahub.metadata.com.linkedin.pegasus2avro.container import ContainerProperties
from datahub.metadata.com.linkedin.pegasus2avro.schema import (
    OtherSchema,
    SchemaField,
    SchemaMetadata,
)
from datahub.metadata.schema_classes import (
    BrowsePathEntryClass,
    BrowsePathsV2Class,
    ContainerClass,
    DataPlatformInstanceClass,
    DatasetPropertiesClass,
    OwnerClass,
    OwnershipClass,
    OwnershipTypeClass,
    TimeStampClass,
    _Aspect,
)
from datahub.utilities.perf_timer import PerfTimer
from datahub.utilities.threaded_iterator_executor import ThreadedIteratorExecutor

LOGGER = logging.getLogger(__name__)
logging.getLogger("azure.core.pipeline.policies.http_logging_policy").setLevel(
    logging.WARNING
)


@platform_name("Iceberg")
@support_status(SupportStatus.TESTING)
@config_class(IcebergSourceConfig)
@capability(
    SourceCapability.PLATFORM_INSTANCE,
    "Optionally enabled via configuration, an Iceberg instance represents the catalog name where the table is stored.",
)
@capability(SourceCapability.DOMAINS, "Currently not supported.", supported=False)
@capability(SourceCapability.DATA_PROFILING, "Optionally enabled via configuration.")
@capability(
    SourceCapability.PARTITION_SUPPORT, "Currently not supported.", supported=False
)
@capability(SourceCapability.DESCRIPTIONS, "Enabled by default.")
@capability(
    SourceCapability.OWNERSHIP,
    "Automatically ingests ownership information from table properties based on `user_ownership_property` and `group_ownership_property`",
)
@capability(SourceCapability.DELETION_DETECTION, "Enabled via stateful ingestion")
class IcebergSource(StatefulIngestionSourceBase):
    """
    ## Integration Details

    The DataHub Iceberg source plugin extracts metadata from [Iceberg tables](https://iceberg.apache.org/spec/) stored in a distributed or local file system.
    Typically, Iceberg tables are stored in a distributed file system like S3 or Azure Data Lake Storage (ADLS) and registered in a catalog.  There are various catalog
    implementations like Filesystem-based, RDBMS-based or even REST-based catalogs.  This Iceberg source plugin relies on the
    [pyiceberg library](https://py.iceberg.apache.org/).
    """

    platform: str = "iceberg"

    def __init__(self, config: IcebergSourceConfig, ctx: PipelineContext) -> None:
        super().__init__(config, ctx)
        self.report: IcebergSourceReport = IcebergSourceReport()
        self.config: IcebergSourceConfig = config
        self.ctx: PipelineContext = ctx

    @classmethod
    def create(cls, config_dict: Dict, ctx: PipelineContext) -> "IcebergSource":
        config = IcebergSourceConfig.parse_obj(config_dict)
        return cls(config, ctx)

    def get_workunit_processors(self) -> List[Optional[MetadataWorkUnitProcessor]]:
        # This source needs to overwrite standard `get_workunit_processor`, because it is unique in terms of usage
        # of parallelism. Because of this, 2 processors won't work as expected:
        # 1. browse_path_processor - it needs aspects for a single entity to be continuous - which is not guaranteed
        #    in this source
        # 2. automatic stamping with systemMetadata - in current implementation of the Source class this processor
        #    would have been applied in a thread (single) shared between the source, processors and transformers.
        #    Since the metadata scraping happens in separate threads, this could lead to difference between
        #    time used by systemMetadata and actual time at which metadata was read
        auto_lowercase_dataset_urns: Optional[MetadataWorkUnitProcessor] = None
        if (
            self.ctx.pipeline_config
            and self.ctx.pipeline_config.source
            and self.ctx.pipeline_config.source.config
            and (
                (
                    hasattr(
                        self.ctx.pipeline_config.source.config,
                        "convert_urns_to_lowercase",
                    )
                    and self.ctx.pipeline_config.source.config.convert_urns_to_lowercase
                )
                or (
                    hasattr(self.ctx.pipeline_config.source.config, "get")
                    and self.ctx.pipeline_config.source.config.get(
                        "convert_urns_to_lowercase"
                    )
                )
            )
        ):
            auto_lowercase_dataset_urns = auto_lowercase_urns

        return [
            auto_lowercase_dataset_urns,
            auto_materialize_referenced_tags_terms,
            partial(
                auto_fix_duplicate_schema_field_paths, platform=self._infer_platform()
            ),
            partial(auto_fix_empty_field_paths, platform=self._infer_platform()),
            partial(auto_workunit_reporter, self.get_report()),
            auto_patch_last_modified,
            EnsureAspectSizeProcessor(self.get_report()).ensure_aspect_size,
            StaleEntityRemovalHandler.create(
                self, self.config, self.ctx
            ).workunit_processor,
        ]

    def _get_namespaces(self, catalog: Catalog) -> Iterable[Identifier]:
        namespaces = catalog.list_namespaces()
        LOGGER.debug(
            f"Retrieved {len(namespaces)} namespaces, first 10: {namespaces[:10]}"
        )
        self.report.report_no_listed_namespaces(len(namespaces))
        for namespace in namespaces:
            namespace_repr = ".".join(namespace)
            if not self.config.namespace_pattern.allowed(namespace_repr):
                LOGGER.info(
                    f"Namespace {namespace_repr} is not allowed by config pattern, skipping"
                )
                self.report.report_dropped(f"{namespace_repr}.*")
                continue
            yield namespace

    def _get_datasets(
        self, catalog: Catalog, namespaces: Iterable[Tuple[Identifier, str]]
    ) -> Iterable[Tuple[Identifier, str]]:
        LOGGER.debug("Starting to retrieve tables")
        tables_count = 0
        for namespace, namespace_urn in namespaces:
            try:
                tables = catalog.list_tables(namespace)
                tables_count += len(tables)
                LOGGER.debug(
                    f"Retrieved {len(tables)} tables for namespace: {namespace}, in total retrieved {tables_count}, first 10: {tables[:10]}"
                )
                self.report.report_listed_tables_for_namespace(
                    ".".join(namespace), len(tables)
                )
                yield from [(table, namespace_urn) for table in tables]
            except NoSuchNamespaceError as e:
                self.report.warning(
                    title="No such namespace",
                    message="Skipping the missing namespace.",
                    context=str(namespace),
                    exc=e,
                )
            except Exception as e:
                self.report.report_failure(
                    title="Error when processing a namespace",
                    message="Skipping the namespace due to errors while processing it.",
                    context=str(namespace),
                    exc=e,
                )

    def get_workunits_internal(self) -> Iterable[MetadataWorkUnit]:
        thread_local = threading.local()

        def _try_processing_dataset(
            dataset_path: Tuple[str, ...], dataset_name: str, namespace_urn: str
        ) -> Iterable[MetadataWorkUnit]:
            try:
                if not hasattr(thread_local, "local_catalog"):
                    LOGGER.debug(
                        f"Didn't find local_catalog in thread_local ({thread_local}), initializing new catalog"
                    )
                    thread_local.local_catalog = self.config.get_catalog()

                if not hasattr(thread_local, "stamping_processor"):
                    LOGGER.debug(
                        f"Didn't find stamping_processor in thread_local ({thread_local}), initializing new workunit processor"
                    )
                    thread_local.stamping_processor = AutoSystemMetadata(self.ctx)

                with PerfTimer() as timer:
                    table = thread_local.local_catalog.load_table(dataset_path)
                    time_taken = timer.elapsed_seconds()
                    self.report.report_table_load_time(
                        time_taken, dataset_name, table.metadata_location
                    )
                LOGGER.debug(f"Loaded table: {table.name()}, time taken: {time_taken}")
                dataset_urn: str = make_dataset_urn_with_platform_instance(
                    self.platform,
                    dataset_name,
                    self.config.platform_instance,
                    self.config.env,
                )
                for aspect in self._create_iceberg_table_aspects(
                    dataset_name, table, namespace_urn
                ):
                    yield thread_local.stamping_processor.stamp_wu(
                        MetadataChangeProposalWrapper(
                            entityUrn=dataset_urn, aspect=aspect
                        ).as_workunit()
                    )
            except NoSuchPropertyException as e:
                self.report.warning(
                    title="Unable to process table",
                    message="Table was not processed due to expected property missing (table is probably not an iceberg table).",
                    context=dataset_name,
                    exc=e,
                )
            except NoSuchIcebergTableError as e:
                self.report.warning(
                    title="Skipped non-iceberg table",
                    message="Table was recognized as non-iceberg and skipped.",
                    context=dataset_name,
                    exc=e,
                )
            except NoSuchTableError as e:
                self.report.warning(
                    title="Table not found",
                    message="Table was returned by the catalog in the list of table but catalog can't find its details, table was skipped.",
                    context=dataset_name,
                    exc=e,
                )
            except FileNotFoundError as e:
                self.report.warning(
                    title="Manifest file not found",
                    message="Couldn't find manifest file to read for the table, skipping it.",
                    context=dataset_name,
                    exc=e,
                )
            except ServerError as e:
                self.report.warning(
                    title="Iceberg REST Server Error",
                    message="Iceberg returned 500 HTTP status when trying to process a table, skipping it.",
                    context=dataset_name,
                    exc=e,
                )
            except ValueError as e:
                if "Could not initialize FileIO" not in str(e):
                    raise
                self.report.warning(
                    title="Could not initialize FileIO",
                    message="Could not initialize FileIO for a table (are you using custom FileIO?). Skipping the table.",
                    context=dataset_name,
                    exc=e,
                )

        def _process_dataset(
            dataset_path: Identifier, namespace_urn: str
        ) -> Iterable[MetadataWorkUnit]:
            try:
                LOGGER.debug(f"Processing dataset for path {dataset_path}")
                dataset_name = ".".join(dataset_path)
                if not self.config.table_pattern.allowed(dataset_name):
                    # Dataset name is rejected by pattern, report as dropped.
                    self.report.report_dropped(dataset_name)
                    LOGGER.debug(
                        f"Skipping table {dataset_name} due to not being allowed by the config pattern"
                    )
                    return

                yield from _try_processing_dataset(
                    dataset_path, dataset_name, namespace_urn
                )
            except Exception as e:
                self.report.report_failure(
                    title="Error when processing a table",
                    message="Skipping the table due to errors when processing it.",
                    context=str(dataset_path),
                    exc=e,
                )

        try:
            catalog = self.config.get_catalog()
        except Exception as e:
            self.report.report_failure(
                title="Failed to initialize catalog object",
                message="Couldn't start the ingestion due to failure to initialize catalog object.",
                exc=e,
            )
            return

        try:
            stamping_processor = AutoSystemMetadata(self.ctx)
            namespace_ids = self._get_namespaces(catalog)
            namespaces: List[Tuple[Identifier, str]] = []
            for namespace in namespace_ids:
                namespace_repr = ".".join(namespace)
                LOGGER.debug(f"Processing namespace {namespace_repr}")
                namespace_urn = make_container_urn(
                    NamespaceKey(
                        namespace=namespace_repr,
                        platform=self.platform,
                        instance=self.config.platform_instance,
                        env=self.config.env,
                    )
                )
                namespace_properties: Properties = catalog.load_namespace_properties(
                    namespace
                )
                namespaces.append((namespace, namespace_urn))
                for aspect in self._create_iceberg_namespace_aspects(
                    namespace, namespace_properties
                ):
                    yield stamping_processor.stamp_wu(
                        MetadataChangeProposalWrapper(
                            entityUrn=namespace_urn, aspect=aspect
                        ).as_workunit()
                    )
            LOGGER.debug("Namespaces ingestion completed")
        except Exception as e:
            self.report.report_failure(
                title="Failed to list namespaces",
                message="Couldn't start the ingestion due to a failure to process the list of the namespaces",
                exc=e,
            )
            return

        for wu in ThreadedIteratorExecutor.process(
            worker_func=_process_dataset,
            args_list=[
                (dataset_path, namespace_urn)
                for dataset_path, namespace_urn in self._get_datasets(
                    catalog, namespaces
                )
            ],
            max_workers=self.config.processing_threads,
        ):
            yield wu

    def _create_iceberg_table_aspects(
        self, dataset_name: str, table: Table, namespace_urn: str
    ) -> Iterable[_Aspect]:
        with PerfTimer() as timer:
            self.report.report_table_scanned(dataset_name)
            LOGGER.debug(f"Processing table {dataset_name}")
            yield Status(removed=False)
            yield SubTypes(typeNames=[DatasetSubTypes.TABLE])

            yield self._get_dataset_properties_aspect(dataset_name, table)

            dataset_ownership = self._get_ownership_aspect(table)
            if dataset_ownership:
                LOGGER.debug(
                    f"Adding ownership: {dataset_ownership} to the dataset {dataset_name}"
                )
                yield dataset_ownership

            yield self._create_schema_metadata(dataset_name, table)
            dpi = self._get_dataplatform_instance_aspect()
            yield dpi
            yield self._create_browse_paths_aspect(dpi.instance, str(namespace_urn))
            yield ContainerClass(container=str(namespace_urn))

        self.report.report_table_processing_time(
            timer.elapsed_seconds(), dataset_name, table.metadata_location
        )

        if self.config.is_profiling_enabled():
            profiler = IcebergProfiler(self.report, self.config.profiling)
            yield from profiler.profile_table(dataset_name, table)

    def _create_browse_paths_aspect(
        self,
        platform_instance_urn: Optional[str] = None,
        container_urn: Optional[str] = None,
    ) -> BrowsePathsV2Class:
        path = []
        if platform_instance_urn:
            path.append(
                BrowsePathEntryClass(
                    id=platform_instance_urn, urn=platform_instance_urn
                )
            )
        if container_urn:
            path.append(BrowsePathEntryClass(id=container_urn, urn=container_urn))
        return BrowsePathsV2Class(path=path)

    def _get_partition_aspect(self, table: Table) -> Optional[str]:
        """Extracts partition information from the provided table and returns a JSON array representing the [partition spec](https://iceberg.apache.org/spec/?#partition-specs) of the table.
        Each element of the returned array represents a field in the [partition spec](https://iceberg.apache.org/spec/?#partition-specs) that follows [Appendix-C](https://iceberg.apache.org/spec/?#appendix-c-json-serialization) of the Iceberg specification.
        Extra information has been added to this spec to make the information more user-friendly.

        Since Datahub does not have a place in its model to store this information, it is saved as a JSON string and displayed as a table property.

        Here is an example:
        ```json
        "partition-spec": "[{\"name\": \"timeperiod_loaded\", \"transform\": \"identity\", \"source\": \"timeperiod_loaded\", \"source-id\": 19, \"source-type\": \"date\", \"field-id\": 1000}]",
        ```

        Args:
            table (Table): The Iceberg table to extract partition spec from.

        Returns:
            str: JSON representation of the partition spec of the provided table (empty array if table is not partitioned) or `None` if an error occured.
        """
        try:
            return json.dumps(
                [
                    {
                        "name": partition.name,
                        "transform": str(partition.transform),
                        "source": str(
                            table.schema().find_column_name(partition.source_id)
                        ),
                        "source-id": partition.source_id,
                        "source-type": str(
                            table.schema().find_type(partition.source_id)
                        ),
                        "field-id": partition.field_id,
                    }
                    for partition in table.spec().fields
                ]
            )
        except Exception as e:
            self.report.warning(
                title="Failed to extract partition information",
                message="Failed to extract partition information for a table. Table metadata will be ingested without it.",
                context=str(table.name),
                exc=e,
            )
            return None

    def _get_dataset_properties_aspect(
        self, dataset_name: str, table: Table
    ) -> DatasetPropertiesClass:
        created: Optional[TimeStampClass] = None
        custom_properties = table.metadata.properties.copy()
        custom_properties["location"] = table.metadata.location
        custom_properties["format-version"] = str(table.metadata.format_version)
        custom_properties["partition-spec"] = str(self._get_partition_aspect(table))
        last_modified: Optional[int] = table.metadata.last_updated_ms
        if table.current_snapshot():
            custom_properties["snapshot-id"] = str(table.current_snapshot().snapshot_id)
            custom_properties["manifest-list"] = table.current_snapshot().manifest_list
            if not last_modified:
                last_modified = int(table.current_snapshot().timestamp_ms)
        if "created-at" in custom_properties:
            try:
                dt = dateutil_parser.isoparse(custom_properties["created-at"])
                created = TimeStampClass(int(dt.timestamp() * 1000))
            except Exception as ex:
                LOGGER.warning(
                    f"Exception while trying to parse creation date {custom_properties['created-at']}, ignoring: {ex}"
                )

        return DatasetPropertiesClass(
            name=table.name()[-1],
            description=table.metadata.properties.get("comment", None),
            customProperties=custom_properties,
            lastModified=TimeStampClass(last_modified)
            if last_modified is not None
            else None,
            created=created,
            qualifiedName=dataset_name,
        )

    def _get_ownership_aspect(self, table: Table) -> Optional[OwnershipClass]:
        owners = []
        if self.config.user_ownership_property:
            if self.config.user_ownership_property in table.metadata.properties:
                user_owner = table.metadata.properties[
                    self.config.user_ownership_property
                ]
                owners.append(
                    OwnerClass(
                        owner=make_user_urn(user_owner),
                        type=OwnershipTypeClass.TECHNICAL_OWNER,
                        source=None,
                    )
                )
        if self.config.group_ownership_property:
            if self.config.group_ownership_property in table.metadata.properties:
                group_owner = table.metadata.properties[
                    self.config.group_ownership_property
                ]
                owners.append(
                    OwnerClass(
                        owner=make_group_urn(group_owner),
                        type=OwnershipTypeClass.TECHNICAL_OWNER,
                        source=None,
                    )
                )
        return OwnershipClass(owners=owners) if owners else None

    def _get_dataplatform_instance_aspect(self) -> DataPlatformInstanceClass:
        return DataPlatformInstanceClass(
            platform=make_data_platform_urn(self.platform),
            instance=make_dataplatform_instance_urn(
                self.platform, self.config.platform_instance
            )
            if self.config.platform_instance
            else None,
        )

    def _create_schema_metadata(
        self, dataset_name: str, table: Table
    ) -> SchemaMetadata:
        schema_fields = self._get_schema_fields_for_schema(table.schema())
        schema_metadata = SchemaMetadata(
            schemaName=dataset_name,
            platform=make_data_platform_urn(self.platform),
            version=0,
            hash="",
            platformSchema=OtherSchema(rawSchema=str(table.schema())),
            fields=schema_fields,
        )
        return schema_metadata

    def _get_schema_fields_for_schema(
        self,
        schema: Schema,
    ) -> List[SchemaField]:
        avro_schema = visit(schema, ToAvroSchemaIcebergVisitor())
        schema_fields = schema_util.avro_schema_to_mce_fields(
            json.dumps(avro_schema), default_nullable=False
        )
        return schema_fields

    def get_report(self) -> SourceReport:
        return self.report

    def _create_iceberg_namespace_aspects(
        self, namespace: Identifier, properties: Properties
    ) -> Iterable[_Aspect]:
        namespace_repr = ".".join(namespace)
        custom_properties: Dict[str, str] = {}
        for k, v in properties.items():
            try:
                custom_properties[str(k)] = str(v)
            except Exception as e:
                LOGGER.warning(
                    f"Exception when trying to parse namespace properties for {namespace_repr}. Exception: {e}"
                )
        yield Status(removed=False)
        yield ContainerProperties(
            name=namespace_repr,
            qualifiedName=namespace_repr,
            env=self.config.env,
            customProperties=custom_properties,
        )
        yield SubTypes(typeNames=[DatasetContainerSubTypes.NAMESPACE])
        dpi = self._get_dataplatform_instance_aspect()
        yield dpi
        yield self._create_browse_paths_aspect(dpi.instance)


class ToAvroSchemaIcebergVisitor(SchemaVisitorPerPrimitiveType[Dict[str, Any]]):
    """Implementation of a visitor to build an Avro schema as a dictionary from an Iceberg schema."""

    @staticmethod
    def _gen_name(prefix: str) -> str:
        return f"{prefix}{str(uuid.uuid4()).replace('-', '')}"

    def schema(self, schema: Schema, struct_result: Dict[str, Any]) -> Dict[str, Any]:
        return struct_result

    def struct(
        self, struct: StructType, field_results: List[Dict[str, Any]]
    ) -> Dict[str, Any]:
        nullable = True
        return {
            "type": "record",
            "name": self._gen_name("__struct_"),
            "fields": field_results,
            "native_data_type": str(struct),
            "_nullable": nullable,
        }

    def field(self, field: NestedField, field_result: Dict[str, Any]) -> Dict[str, Any]:
        field_result["_nullable"] = not field.required
        return {
            "name": field.name,
            "type": field_result,
            "doc": field.doc,
        }

    def list(
        self, list_type: ListType, element_result: Dict[str, Any]
    ) -> Dict[str, Any]:
        return {
            "type": "array",
            "items": element_result,
            "native_data_type": str(list_type),
            "_nullable": not list_type.element_required,
        }

    def map(
        self,
        map_type: MapType,
        key_result: Dict[str, Any],
        value_result: Dict[str, Any],
    ) -> Dict[str, Any]:
        # The Iceberg Map type will be handled differently.  The idea is to translate the map
        # similar to the Map.Entry struct of Java i.e. as an array of map_entry struct, where
        # the map_entry struct has a key field and a value field. The key and value type can
        # be complex or primitive types.
        key_result["_nullable"] = False
        value_result["_nullable"] = not map_type.value_required
        map_entry = {
            "type": "record",
            "name": self._gen_name("__map_entry_"),
            "fields": [
                {
                    "name": "key",
                    "type": key_result,
                },
                {
                    "name": "value",
                    "type": value_result,
                },
            ],
        }
        return {
            "type": "array",
            "items": map_entry,
            "native_data_type": str(map_type),
        }

    def visit_fixed(self, fixed_type: FixedType) -> Dict[str, Any]:
        return {
            "type": "fixed",
            "name": self._gen_name("__fixed_"),
            "size": len(fixed_type),
            "native_data_type": str(fixed_type),
        }

    def visit_decimal(self, decimal_type: DecimalType) -> Dict[str, Any]:
        # Also of interest: https://avro.apache.org/docs/current/spec.html#Decimal
        return {
            # "type": "bytes", # when using bytes, avro drops _nullable attribute and others.  See unit test.
            "type": "fixed",  # to fix avro bug ^ resolved by using a fixed type
            "name": self._gen_name(
                "__fixed_"
            ),  # to fix avro bug ^ resolved by using a fixed type
            "size": 1,  # to fix avro bug ^ resolved by using a fixed type
            "logicalType": "decimal",
            "precision": decimal_type.precision,
            "scale": decimal_type.scale,
            "native_data_type": str(decimal_type),
        }

    def visit_boolean(self, boolean_type: BooleanType) -> Dict[str, Any]:
        return {
            "type": "boolean",
            "native_data_type": str(boolean_type),
        }

    def visit_integer(self, integer_type: IntegerType) -> Dict[str, Any]:
        return {
            "type": "int",
            "native_data_type": str(integer_type),
        }

    def visit_long(self, long_type: LongType) -> Dict[str, Any]:
        return {
            "type": "long",
            "native_data_type": str(long_type),
        }

    def visit_float(self, float_type: FloatType) -> Dict[str, Any]:
        return {
            "type": "float",
            "native_data_type": str(float_type),
        }

    def visit_double(self, double_type: DoubleType) -> Dict[str, Any]:
        return {
            "type": "double",
            "native_data_type": str(double_type),
        }

    def visit_date(self, date_type: DateType) -> Dict[str, Any]:
        return {
            "type": "int",
            "logicalType": "date",
            "native_data_type": str(date_type),
        }

    def visit_time(self, time_type: TimeType) -> Dict[str, Any]:
        return {
            "type": "long",
            "logicalType": "time-micros",
            "native_data_type": str(time_type),
        }

    def visit_timestamp(self, timestamp_type: TimestampType) -> Dict[str, Any]:
        # Avro supports 2 types of timestamp:
        #  - Timestamp: independent of a particular timezone or calendar (TZ information is lost)
        #  - Local Timestamp: represents a timestamp in a local timezone, regardless of what specific time zone is considered local
        # utcAdjustment: bool = True
        return {
            "type": "long",
            "logicalType": "timestamp-micros",
            # Commented out since Avro's Python implementation (1.11.0) does not support local-timestamp-micros, even though it exists in the spec.
            # See bug report: https://issues.apache.org/jira/browse/AVRO-3476 and PR https://github.com/apache/avro/pull/1634
            # "logicalType": "timestamp-micros"
            # if timestamp_type.adjust_to_utc
            # else "local-timestamp-micros",
            "native_data_type": str(timestamp_type),
        }

    # visit_timestamptz() is required when using pyiceberg >= 0.5.0, which is essentially a duplicate
    # of visit_timestampz().  The function has been renamed from visit_timestampz().
    # Once Datahub can upgrade its pyiceberg dependency to >=0.5.0, the visit_timestampz() function can be safely removed.
    def visit_timestamptz(self, timestamptz_type: TimestamptzType) -> Dict[str, Any]:
        # Avro supports 2 types of timestamp:
        #  - Timestamp: independent of a particular timezone or calendar (TZ information is lost)
        #  - Local Timestamp: represents a timestamp in a local timezone, regardless of what specific time zone is considered local
        # utcAdjustment: bool = True
        return {
            "type": "long",
            "logicalType": "timestamp-micros",
            # Commented out since Avro's Python implementation (1.11.0) does not support local-timestamp-micros, even though it exists in the spec.
            # See bug report: https://issues.apache.org/jira/browse/AVRO-3476 and PR https://github.com/apache/avro/pull/1634
            # "logicalType": "timestamp-micros"
            # if timestamp_type.adjust_to_utc
            # else "local-timestamp-micros",
            "native_data_type": str(timestamptz_type),
        }

    def visit_timestampz(self, timestamptz_type: TimestamptzType) -> Dict[str, Any]:
        # Avro supports 2 types of timestamp:
        #  - Timestamp: independent of a particular timezone or calendar (TZ information is lost)
        #  - Local Timestamp: represents a timestamp in a local timezone, regardless of what specific time zone is considered local
        # utcAdjustment: bool = True
        return {
            "type": "long",
            "logicalType": "timestamp-micros",
            # Commented out since Avro's Python implementation (1.11.0) does not support local-timestamp-micros, even though it exists in the spec.
            # See bug report: https://issues.apache.org/jira/browse/AVRO-3476 and PR https://github.com/apache/avro/pull/1634
            # "logicalType": "timestamp-micros"
            # if timestamp_type.adjust_to_utc
            # else "local-timestamp-micros",
            "native_data_type": str(timestamptz_type),
        }

    def visit_string(self, string_type: StringType) -> Dict[str, Any]:
        return {
            "type": "string",
            "native_data_type": str(string_type),
        }

    def visit_uuid(self, uuid_type: UUIDType) -> Dict[str, Any]:
        return {
            "type": "string",
            "logicalType": "uuid",
            "native_data_type": str(uuid_type),
        }

    def visit_binary(self, binary_type: BinaryType) -> Dict[str, Any]:
        return {
            "type": "bytes",
            "native_data_type": str(binary_type),
        }



import contextlib
import dataclasses
import enum
import functools
import json
import logging
import os
import pathlib
import tempfile
import uuid
from collections import defaultdict
from datetime import datetime, timezone
from typing import Callable, Dict, Iterable, List, Optional, Set, Union, cast

import datahub.emitter.mce_builder as builder
import datahub.metadata.schema_classes as models
from datahub.configuration.time_window_config import get_time_bucket
from datahub.emitter.mce_builder import get_sys_time, make_ts_millis
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.emitter.sql_parsing_builder import compute_upstream_fields
from datahub.ingestion.api.closeable import Closeable
from datahub.ingestion.api.report import Report
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.graph.client import DataHubGraph
from datahub.ingestion.source.usage.usage_common import BaseUsageConfig, UsageAggregator
from datahub.metadata.urns import (
    CorpGroupUrn,
    CorpUserUrn,
    DataPlatformUrn,
    DatasetUrn,
    QueryUrn,
    SchemaFieldUrn,
    Urn,
)
from datahub.sql_parsing.fingerprint_utils import generate_hash
from datahub.sql_parsing.schema_resolver import (
    SchemaResolver,
    SchemaResolverInterface,
    _SchemaResolverWithExtras,
)
from datahub.sql_parsing.sql_parsing_common import QueryType, QueryTypeProps
from datahub.sql_parsing.sqlglot_lineage import (
    ColumnLineageInfo,
    ColumnRef,
    DownstreamColumnRef,
    SqlParsingResult,
    _sqlglot_lineage_cached,
    infer_output_schema,
    sqlglot_lineage,
)
from datahub.sql_parsing.sqlglot_utils import (
    _parse_statement,
    get_query_fingerprint,
    try_format_query,
)
from datahub.sql_parsing.tool_meta_extractor import (
    ToolMetaExtractor,
    ToolMetaExtractorReport,
)
from datahub.utilities.cooperative_timeout import CooperativeTimeoutError
from datahub.utilities.file_backed_collections import (
    ConnectionWrapper,
    FileBackedDict,
    FileBackedList,
)
from datahub.utilities.groupby import groupby_unsorted
from datahub.utilities.lossy_collections import LossyDict, LossyList
from datahub.utilities.ordered_set import OrderedSet
from datahub.utilities.perf_timer import PerfTimer

logger = logging.getLogger(__name__)
QueryId = str
UrnStr = str


class QueryLogSetting(enum.Enum):
    DISABLED = "DISABLED"
    STORE_ALL = "STORE_ALL"
    STORE_FAILED = "STORE_FAILED"


_DEFAULT_USER_URN = CorpUserUrn("_ingestion")
_MISSING_SESSION_ID = "__MISSING_SESSION_ID"
_DEFAULT_QUERY_LOG_SETTING = QueryLogSetting[
    os.getenv("DATAHUB_SQL_AGG_QUERY_LOG") or QueryLogSetting.DISABLED.name
]
MAX_UPSTREAM_TABLES_COUNT = 300
MAX_FINEGRAINEDLINEAGE_COUNT = 2000


@dataclasses.dataclass
class LoggedQuery:
    query: str
    session_id: Optional[str]
    timestamp: Optional[datetime]
    user: Optional[UrnStr]
    default_db: Optional[str]
    default_schema: Optional[str]


@dataclasses.dataclass
class ObservedQuery:
    query: str
    session_id: Optional[str] = None
    timestamp: Optional[datetime] = None
    user: Optional[Union[CorpUserUrn, CorpGroupUrn]] = None
    default_db: Optional[str] = None
    default_schema: Optional[str] = None
    query_hash: Optional[str] = None
    usage_multiplier: int = 1

    # Use this to store addtitional key-value information about query for debugging
    extra_info: Optional[dict] = None


@dataclasses.dataclass
class ViewDefinition:
    view_definition: str
    default_db: Optional[str] = None
    default_schema: Optional[str] = None


@dataclasses.dataclass
class QueryMetadata:
    query_id: QueryId

    # raw_query_string: str
    formatted_query_string: str

    session_id: str  # will be _MISSING_SESSION_ID if not present
    query_type: QueryType
    lineage_type: str  # from models.DatasetLineageTypeClass
    latest_timestamp: Optional[datetime]
    actor: Optional[Union[CorpUserUrn, CorpGroupUrn]]

    upstreams: List[UrnStr]  # this is direct upstreams, which may be temp tables
    column_lineage: List[ColumnLineageInfo]
    column_usage: Dict[UrnStr, Set[UrnStr]]  # TODO: Change to an OrderedSet
    confidence_score: float

    used_temp_tables: bool = True

    origin: Optional[Urn] = None

    def make_created_audit_stamp(self) -> models.AuditStampClass:
        return models.AuditStampClass(
            time=make_ts_millis(self.latest_timestamp) or 0,
            actor=(self.actor or _DEFAULT_USER_URN).urn(),
        )

    def make_last_modified_audit_stamp(self) -> models.AuditStampClass:
        return models.AuditStampClass(
            time=make_ts_millis(self.latest_timestamp or datetime.now(tz=timezone.utc))
            or 0,
            actor=(self.actor or _DEFAULT_USER_URN).urn(),
        )

    def get_subjects(
        self,
        downstream_urn: Optional[str],
        include_fields: bool,
    ) -> List[UrnStr]:
        query_subject_urns = OrderedSet[UrnStr]()
        for upstream in self.upstreams:
            query_subject_urns.add(upstream)
            if include_fields:
                for column in sorted(self.column_usage.get(upstream, [])):
                    query_subject_urns.add(
                        builder.make_schema_field_urn(upstream, column)
                    )
        if downstream_urn:
            query_subject_urns.add(downstream_urn)
            if include_fields:
                for column_lineage in self.column_lineage:
                    query_subject_urns.add(
                        builder.make_schema_field_urn(
                            downstream_urn, column_lineage.downstream.column
                        )
                    )
        return list(query_subject_urns)

    def make_query_properties(self) -> models.QueryPropertiesClass:
        return models.QueryPropertiesClass(
            statement=models.QueryStatementClass(
                value=self.formatted_query_string,
                language=models.QueryLanguageClass.SQL,
            ),
            source=models.QuerySourceClass.SYSTEM,
            created=self.make_created_audit_stamp(),
            lastModified=self.make_last_modified_audit_stamp(),
        )


def make_query_subjects(urns: List[UrnStr]) -> models.QuerySubjectsClass:
    return models.QuerySubjectsClass(
        subjects=[models.QuerySubjectClass(entity=urn) for urn in urns]
    )


@dataclasses.dataclass
class KnownQueryLineageInfo:
    query_text: str

    downstream: UrnStr
    upstreams: List[UrnStr]
    column_lineage: Optional[List[ColumnLineageInfo]] = None
    column_usage: Optional[Dict[UrnStr, Set[UrnStr]]] = None

    timestamp: Optional[datetime] = None
    session_id: Optional[str] = None
    query_type: QueryType = QueryType.UNKNOWN
    query_id: Optional[str] = None


@dataclasses.dataclass
class KnownLineageMapping:
    upstream_urn: UrnStr
    downstream_urn: UrnStr
    lineage_type: str = models.DatasetLineageTypeClass.COPY


@dataclasses.dataclass
class TableRename:
    original_urn: UrnStr
    new_urn: UrnStr
    query: Optional[str] = None
    session_id: str = _MISSING_SESSION_ID
    timestamp: Optional[datetime] = None


@dataclasses.dataclass
class TableSwap:
    urn1: UrnStr
    urn2: UrnStr
    query: Optional[str] = None
    session_id: str = _MISSING_SESSION_ID
    timestamp: Optional[datetime] = None

    def id(self) -> str:
        # TableSwap(A,B) is same as TableSwap(B,A)
        return str(hash(frozenset([self.urn1, self.urn2])))


@dataclasses.dataclass
class PreparsedQuery:
    # If not provided, we will generate one using the fingerprint generator.
    query_id: Optional[QueryId]

    query_text: str

    upstreams: List[UrnStr]
    downstream: Optional[UrnStr] = None
    column_lineage: Optional[List[ColumnLineageInfo]] = None
    column_usage: Optional[Dict[UrnStr, Set[UrnStr]]] = None
    inferred_schema: Optional[List[models.SchemaFieldClass]] = None
    confidence_score: float = 1.0

    query_count: int = 1
    user: Optional[Union[CorpUserUrn, CorpGroupUrn]] = None
    timestamp: Optional[datetime] = None
    session_id: str = _MISSING_SESSION_ID
    query_type: QueryType = QueryType.UNKNOWN
    query_type_props: QueryTypeProps = dataclasses.field(
        default_factory=lambda: QueryTypeProps()
    )
    # Use this to store addtitional key-value information about query for debugging
    extra_info: Optional[dict] = None
    origin: Optional[Urn] = None


@dataclasses.dataclass
class SqlAggregatorReport(Report):
    _aggregator: "SqlParsingAggregator"
    query_log_path: Optional[str] = None

    # Observed queries.
    num_observed_queries: int = 0
    num_observed_queries_failed: int = 0
    num_observed_queries_column_timeout: int = 0
    num_observed_queries_column_failed: int = 0
    observed_query_parse_failures: LossyList[str] = dataclasses.field(
        default_factory=LossyList
    )

    # Views.
    num_view_definitions: int = 0
    num_views_failed: int = 0
    num_views_column_timeout: int = 0
    num_views_column_failed: int = 0
    views_parse_failures: LossyDict[UrnStr, str] = dataclasses.field(
        default_factory=LossyDict
    )

    # SQL parsing (over all invocations).
    num_sql_parsed: int = 0
    sql_parsing_timer: PerfTimer = dataclasses.field(default_factory=PerfTimer)
    sql_fingerprinting_timer: PerfTimer = dataclasses.field(default_factory=PerfTimer)
    sql_formatting_timer: PerfTimer = dataclasses.field(default_factory=PerfTimer)
    sql_parsing_cache_stats: Optional[dict] = dataclasses.field(default=None)
    parse_statement_cache_stats: Optional[dict] = dataclasses.field(default=None)
    format_query_cache_stats: Optional[dict] = dataclasses.field(default=None)

    # Other lineage loading metrics.
    num_known_query_lineage: int = 0
    num_preparsed_queries: int = 0
    num_known_mapping_lineage: int = 0
    num_table_renames: int = 0
    num_table_swaps: int = 0

    # Temp tables.
    num_temp_sessions: Optional[int] = None
    num_inferred_temp_schemas: Optional[int] = None
    num_queries_with_temp_tables_in_session: int = 0
    queries_with_temp_upstreams: LossyDict[QueryId, LossyList] = dataclasses.field(
        default_factory=LossyDict
    )
    queries_with_non_authoritative_session: LossyList[QueryId] = dataclasses.field(
        default_factory=LossyList
    )
    make_schema_resolver_timer: PerfTimer = dataclasses.field(default_factory=PerfTimer)

    # Lineage-related.
    schema_resolver_count: Optional[int] = None
    num_unique_query_fingerprints: Optional[int] = None
    num_urns_with_lineage: Optional[int] = None
    num_lineage_skipped_due_to_filters: int = 0
    num_table_lineage_trimmed_due_to_large_size: int = 0
    num_column_lineage_trimmed_due_to_large_size: int = 0

    # Queries.
    num_queries_entities_generated: int = 0
    num_queries_used_in_lineage: Optional[int] = None
    num_queries_skipped_due_to_filters: int = 0

    # Usage-related.
    usage_skipped_missing_timestamp: int = 0
    num_query_usage_stats_generated: int = 0
    num_query_usage_stats_outside_window: int = 0

    # Operation-related.
    num_operations_generated: int = 0
    num_operations_skipped_due_to_filters: int = 0

    # Tool Metadata Extraction
    tool_meta_report: Optional[ToolMetaExtractorReport] = None

    def compute_stats(self) -> None:
        if self._aggregator._closed:
            return
        self.schema_resolver_count = self._aggregator._schema_resolver.schema_count()
        self.num_unique_query_fingerprints = len(self._aggregator._query_map)

        self.num_urns_with_lineage = len(self._aggregator._lineage_map)
        self.num_temp_sessions = len(self._aggregator._temp_lineage_map)
        self.num_inferred_temp_schemas = len(self._aggregator._inferred_temp_schemas)

        self.sql_parsing_cache_stats = _sqlglot_lineage_cached.cache_info()._asdict()
        self.parse_statement_cache_stats = _parse_statement.cache_info()._asdict()
        self.format_query_cache_stats = try_format_query.cache_info()._asdict()

        return super().compute_stats()


class SqlParsingAggregator(Closeable):
    def __init__(
        self,
        *,
        platform: str,
        platform_instance: Optional[str] = None,
        env: str = builder.DEFAULT_ENV,
        schema_resolver: Optional[SchemaResolver] = None,
        graph: Optional[DataHubGraph] = None,
        eager_graph_load: bool = True,
        generate_lineage: bool = True,
        generate_queries: bool = True,
        generate_query_subject_fields: bool = True,
        generate_usage_statistics: bool = False,
        generate_query_usage_statistics: bool = False,
        generate_operations: bool = False,
        usage_config: Optional[BaseUsageConfig] = None,
        is_temp_table: Optional[Callable[[str], bool]] = None,
        is_allowed_table: Optional[Callable[[str], bool]] = None,
        format_queries: bool = True,
        query_log: QueryLogSetting = _DEFAULT_QUERY_LOG_SETTING,
    ) -> None:
        self.platform = DataPlatformUrn(platform)
        self.platform_instance = platform_instance
        self.env = env

        self.generate_lineage = generate_lineage
        self.generate_queries = generate_queries
        self.generate_query_subject_fields = generate_query_subject_fields
        self.generate_usage_statistics = generate_usage_statistics
        self.generate_query_usage_statistics = generate_query_usage_statistics
        self.generate_operations = generate_operations
        if self.generate_queries and not (
            self.generate_lineage or self.generate_query_usage_statistics
        ):
            logger.warning(
                "Queries will not be generated, as neither lineage nor query usage statistics are enabled"
            )

        self.usage_config = usage_config
        if (
            self.generate_usage_statistics or self.generate_query_usage_statistics
        ) and self.usage_config is None:
            raise ValueError("Usage statistics generation requires a usage config")

        self.report = SqlAggregatorReport(_aggregator=self)

        # can be used by BQ where we have a "temp_table_dataset_prefix"
        self._is_temp_table = is_temp_table
        self._is_allowed_table = is_allowed_table

        self.format_queries = format_queries
        self.query_log = query_log

        # The exit stack helps ensure that we close all the resources we open.
        self._exit_stack = contextlib.ExitStack()
        self._closed: bool = False

        # Set up the schema resolver.
        self._schema_resolver: SchemaResolver
        if schema_resolver is not None:
            # If explicitly provided, use it.
            assert self.platform.platform_name == schema_resolver.platform
            assert self.platform_instance == schema_resolver.platform_instance
            assert self.env == schema_resolver.env
            self._schema_resolver = schema_resolver
        elif graph is not None and eager_graph_load and self._need_schemas:
            # Bulk load schemas using the graph client.
            self._schema_resolver = graph.initialize_schema_resolver_from_datahub(
                platform=self.platform.urn(),
                platform_instance=self.platform_instance,
                env=self.env,
            )
        else:
            # Otherwise, use a lazy-loading schema resolver.
            self._schema_resolver = self._exit_stack.enter_context(
                SchemaResolver(
                    platform=self.platform.platform_name,
                    platform_instance=self.platform_instance,
                    env=self.env,
                    graph=graph,
                )
            )
        # Schema resolver for special case (_MISSING_SESSION_ID)
        # This is particularly useful for via temp table lineage if session id is not available.
        self._missing_session_schema_resolver = _SchemaResolverWithExtras(
            base_resolver=self._schema_resolver, extra_schemas={}
        )

        # Initialize internal data structures.
        # This leans pretty heavily on the our query fingerprinting capabilities.
        # In particular, it must be true that if two queries have the same fingerprint,
        # they must generate the same lineage.

        self._shared_connection: Optional[ConnectionWrapper] = None
        if self.query_log != QueryLogSetting.DISABLED:
            # Initialize and log a file to store the queries.
            query_log_path = pathlib.Path(tempfile.mkdtemp()) / "query_log.db"
            self.report.query_log_path = str(query_log_path)

            # By providing a filename explicitly here, we also ensure that the file
            # is not automatically deleted on exit.
            self._shared_connection = self._exit_stack.enter_context(
                ConnectionWrapper(filename=query_log_path)
            )

        # Stores the logged queries.
        self._logged_queries = FileBackedList[LoggedQuery](
            shared_connection=self._shared_connection, tablename="stored_queries"
        )
        self._exit_stack.push(self._logged_queries)

        # Map of query_id -> QueryMetadata
        self._query_map = FileBackedDict[QueryMetadata](
            shared_connection=self._shared_connection, tablename="query_map"
        )
        self._exit_stack.push(self._query_map)

        # Map of downstream urn -> { query ids }
        self._lineage_map = FileBackedDict[OrderedSet[QueryId]](
            shared_connection=self._shared_connection, tablename="lineage_map"
        )
        self._exit_stack.push(self._lineage_map)

        # Map of view urn -> view definition
        self._view_definitions = FileBackedDict[ViewDefinition](
            shared_connection=self._shared_connection, tablename="view_definitions"
        )
        self._exit_stack.push(self._view_definitions)

        # Map of session ID -> {temp table name -> query id}
        # Needs to use the query_map to find the info about the query.
        self._temp_lineage_map = FileBackedDict[Dict[UrnStr, OrderedSet[QueryId]]](
            shared_connection=self._shared_connection, tablename="temp_lineage_map"
        )
        self._exit_stack.push(self._temp_lineage_map)

        # Map of query ID -> schema fields, only for query IDs that generate temp tables.
        self._inferred_temp_schemas = FileBackedDict[List[models.SchemaFieldClass]](
            shared_connection=self._shared_connection,
            tablename="inferred_temp_schemas",
        )
        self._exit_stack.push(self._inferred_temp_schemas)

        # Map of table renames, from original UrnStr to new UrnStr.
        self._table_renames = FileBackedDict[UrnStr](
            shared_connection=self._shared_connection, tablename="table_renames"
        )
        self._exit_stack.push(self._table_renames)

        # Map of table swaps, from unique swap id to TableSwap
        self._table_swaps = FileBackedDict[TableSwap](
            shared_connection=self._shared_connection, tablename="table_swaps"
        )
        self._exit_stack.push(self._table_swaps)

        # Usage aggregator. This will only be initialized if usage statistics are enabled.
        # TODO: Replace with FileBackedDict.
        # TODO: The BaseUsageConfig class is much too broad for our purposes, and has a number of
        # configs that won't be respected here. Using it is misleading.
        self._usage_aggregator: Optional[UsageAggregator[UrnStr]] = None
        if self.generate_usage_statistics:
            assert self.usage_config is not None
            self._usage_aggregator = UsageAggregator(config=self.usage_config)

        # Query usage aggregator.
        # Map of query ID -> { bucket -> count }
        self._query_usage_counts: Optional[FileBackedDict[Dict[datetime, int]]] = None
        if self.generate_query_usage_statistics:
            self._query_usage_counts = FileBackedDict[Dict[datetime, int]](
                shared_connection=self._shared_connection,
                tablename="query_usage_counts",
            )
            self._exit_stack.push(self._query_usage_counts)

        # Tool Extractor
        self._tool_meta_extractor = ToolMetaExtractor.create(graph)
        self.report.tool_meta_report = self._tool_meta_extractor.report

    def close(self) -> None:
        # Compute stats once before closing connections
        self.report.compute_stats()
        self._closed = True
        self._exit_stack.close()

    @property
    def _need_schemas(self) -> bool:
        # Unless the aggregator is totally disabled, we will need schema information.
        return (
            self.generate_lineage
            or self.generate_usage_statistics
            or self.generate_queries
            or self.generate_operations
            or self.generate_query_usage_statistics
        )

    def register_schema(
        self, urn: Union[str, DatasetUrn], schema: models.SchemaMetadataClass
    ) -> None:
        # If lineage or usage is enabled, adds the schema to the schema resolver
        # by putting the condition in here, we can avoid all the conditional
        # logic that we previously needed in each source

        if self._need_schemas:
            self._schema_resolver.add_schema_metadata(str(urn), schema)

    def register_schemas_from_stream(
        self, stream: Iterable[MetadataWorkUnit]
    ) -> Iterable[MetadataWorkUnit]:
        for wu in stream:
            schema_metadata = wu.get_aspect_of_type(models.SchemaMetadataClass)
            if schema_metadata:
                self.register_schema(wu.get_urn(), schema_metadata)

            yield wu

    def _maybe_format_query(self, query: str) -> str:
        if self.format_queries:
            with self.report.sql_formatting_timer:
                return try_format_query(query, self.platform.platform_name)
        return query

    @functools.lru_cache(maxsize=128)
    def _name_from_urn(self, urn: UrnStr) -> Optional[str]:
        urn_obj = DatasetUrn.from_string(urn)
        if urn_obj.platform != self.platform.urn():
            # If this is external (e.g. s3), we don't know the name.
            return None

        name = urn_obj.name
        if (
            platform_instance := self._schema_resolver.platform_instance
        ) and name.startswith(platform_instance):
            # Remove the platform instance from the name.
            name = name[len(platform_instance) + 1 :]
        return name

    def is_temp_table(self, urn: UrnStr) -> bool:
        if self._is_temp_table is None:
            return False
        name = self._name_from_urn(urn)
        if name is None:
            # External tables are not temp tables.
            return False
        return self._is_temp_table(name)

    def is_allowed_table(self, urn: UrnStr, allow_external: bool = True) -> bool:
        if self.is_temp_table(urn):
            return False
        if self._is_allowed_table is None:
            return True
        name = self._name_from_urn(urn)
        if name is None:
            # Treat external tables specially.
            return allow_external
        return self._is_allowed_table(name)

    def add(
        self,
        item: Union[
            KnownQueryLineageInfo,
            KnownLineageMapping,
            PreparsedQuery,
            ObservedQuery,
            TableRename,
            TableSwap,
        ],
    ) -> None:
        if isinstance(item, KnownQueryLineageInfo):
            self.add_known_query_lineage(item)
        elif isinstance(item, KnownLineageMapping):
            self.add_known_lineage_mapping(item.upstream_urn, item.downstream_urn)
        elif isinstance(item, PreparsedQuery):
            self.add_preparsed_query(item)
        elif isinstance(item, ObservedQuery):
            self.add_observed_query(item)
        elif isinstance(item, TableRename):
            self.add_table_rename(item)
        elif isinstance(item, TableSwap):
            self.add_table_swap(item)
        else:
            raise ValueError(f"Cannot add unknown item type: {type(item)}")

    def add_known_query_lineage(
        self, known_query_lineage: KnownQueryLineageInfo, merge_lineage: bool = False
    ) -> None:
        """Add a query and it's precomputed lineage to the aggregator.

        This is useful for cases where we have lineage information that was
        computed outside of the SQL parsing aggregator, e.g. from a data
        warehouse's system tables.

        This will also generate an operation aspect for the query if there is
        a timestamp and the query type field is set to a mutation type.

        Args:
            known_query_lineage: The known query lineage information.
            merge_lineage: Whether to merge the lineage with any existing lineage
                for the query ID.
        """

        self.report.num_known_query_lineage += 1

        # Generate a fingerprint for the query.
        query_fingerprint = known_query_lineage.query_id
        if not query_fingerprint:
            with self.report.sql_fingerprinting_timer:
                query_fingerprint = get_query_fingerprint(
                    known_query_lineage.query_text,
                    platform=self.platform.platform_name,
                )
        formatted_query = self._maybe_format_query(known_query_lineage.query_text)

        # Register the query.
        self._add_to_query_map(
            QueryMetadata(
                query_id=query_fingerprint,
                formatted_query_string=formatted_query,
                session_id=known_query_lineage.session_id or _MISSING_SESSION_ID,
                query_type=known_query_lineage.query_type,
                lineage_type=models.DatasetLineageTypeClass.TRANSFORMED,
                latest_timestamp=known_query_lineage.timestamp,
                actor=None,
                upstreams=known_query_lineage.upstreams,
                column_lineage=known_query_lineage.column_lineage or [],
                column_usage=known_query_lineage.column_usage or {},
                confidence_score=1.0,
            ),
            merge_lineage=merge_lineage,
        )

        # Register the lineage.
        self._lineage_map.for_mutation(
            known_query_lineage.downstream, OrderedSet()
        ).add(query_fingerprint)

    def add_known_lineage_mapping(
        self,
        upstream_urn: UrnStr,
        downstream_urn: UrnStr,
        lineage_type: str = models.DatasetLineageTypeClass.COPY,
    ) -> None:
        """Add a known lineage mapping to the aggregator.

        By mapping, we mean that the downstream is effectively a copy or
        alias of the upstream. This is useful for things like external tables
        (e.g. Redshift Spectrum, Redshift UNLOADs, Snowflake external tables).

        Because this method takes in urns, it does not require that the urns
        are part of the platform that the aggregator is configured for.

        The known lineage mapping does not contribute to usage statistics or operations.

        Args:
            upstream_urn: The upstream dataset URN.
            downstream_urn: The downstream dataset URN.
        """
        logger.debug(
            f"Adding lineage to the map, downstream: {downstream_urn}, upstream: {upstream_urn}"
        )
        self.report.num_known_mapping_lineage += 1

        # We generate a fake "query" object to hold the lineage.
        query_id = self._known_lineage_query_id()

        # Generate CLL if schema of downstream is known
        column_lineage: List[ColumnLineageInfo] = (
            self._generate_identity_column_lineage(
                upstream_urn=upstream_urn, downstream_urn=downstream_urn
            )
        )

        # Register the query.
        self._add_to_query_map(
            QueryMetadata(
                query_id=query_id,
                formatted_query_string="-skip-",
                session_id=_MISSING_SESSION_ID,
                query_type=QueryType.UNKNOWN,
                lineage_type=lineage_type,
                latest_timestamp=None,
                actor=None,
                upstreams=[upstream_urn],
                column_lineage=column_lineage,
                column_usage={},
                confidence_score=1.0,
            )
        )

        # Register the lineage.
        self._lineage_map.for_mutation(downstream_urn, OrderedSet()).add(query_id)

    def _generate_identity_column_lineage(
        self, *, upstream_urn: UrnStr, downstream_urn: UrnStr
    ) -> List[ColumnLineageInfo]:
        column_lineage: List[ColumnLineageInfo] = []
        if self._schema_resolver.has_urn(downstream_urn):
            schema = self._schema_resolver._resolve_schema_info(downstream_urn)
            if schema:
                column_lineage = [
                    ColumnLineageInfo(
                        downstream=DownstreamColumnRef(
                            table=downstream_urn, column=field_path
                        ),
                        upstreams=[ColumnRef(table=upstream_urn, column=field_path)],
                    )
                    for field_path in schema
                ]

        return column_lineage

    def add_view_definition(
        self,
        view_urn: Union[DatasetUrn, UrnStr],
        view_definition: str,
        default_db: Optional[str] = None,
        default_schema: Optional[str] = None,
    ) -> None:
        """Add a view definition to the aggregator.

        View definitions always contribute to lineage, but do not count towards
        usage statistics or operations.

        The actual processing of view definitions is deferred until output time,
        since all schemas will be registered at that point.
        """

        self.report.num_view_definitions += 1

        self._view_definitions[str(view_urn)] = ViewDefinition(
            view_definition=view_definition,
            default_db=default_db,
            default_schema=default_schema,
        )

    def add_observed_query(
        self,
        observed: ObservedQuery,
        is_known_temp_table: bool = False,
        require_out_table_schema: bool = False,
    ) -> None:
        """Add an observed query to the aggregator.

        This will always generate usage. If it's a mutation query, it will also generate
        lineage and an operation. If it's a temp table, the lineage gets logged in a separate
        map, which will get used in subsequent queries with the same session ID.

        This assumes that queries come in order of increasing timestamps.
        """
        self.report.num_observed_queries += 1

        # All queries with no session ID are assumed to be part of the same session.
        session_id = observed.session_id or _MISSING_SESSION_ID

        with self.report.make_schema_resolver_timer:
            # Load in the temp tables for this session.
            schema_resolver: SchemaResolverInterface = (
                self._make_schema_resolver_for_session(session_id)
            )
            session_has_temp_tables = schema_resolver.includes_temp_tables()

        # Run the SQL parser.
        parsed = self._run_sql_parser(
            observed.query,
            default_db=observed.default_db,
            default_schema=observed.default_schema,
            schema_resolver=schema_resolver,
            session_id=session_id,
            timestamp=observed.timestamp,
            user=observed.user,
        )
        if parsed.debug_info.error:
            self.report.observed_query_parse_failures.append(
                f"{parsed.debug_info.error} on query: {observed.query[:100]}"
            )
        if parsed.debug_info.table_error:
            self.report.num_observed_queries_failed += 1
            return  # we can't do anything with this query
        elif parsed.debug_info.column_error:
            self.report.num_observed_queries_column_failed += 1
            if isinstance(parsed.debug_info.column_error, CooperativeTimeoutError):
                self.report.num_observed_queries_column_timeout += 1

        query_fingerprint = observed.query_hash or parsed.query_fingerprint
        self.add_preparsed_query(
            PreparsedQuery(
                query_id=query_fingerprint,
                query_text=observed.query,
                query_count=observed.usage_multiplier,
                timestamp=observed.timestamp,
                user=observed.user,
                session_id=session_id,
                query_type=parsed.query_type,
                query_type_props=parsed.query_type_props,
                upstreams=parsed.in_tables,
                downstream=parsed.out_tables[0] if parsed.out_tables else None,
                column_lineage=parsed.column_lineage,
                # TODO: We need a full list of columns referenced, not just the out tables.
                column_usage=compute_upstream_fields(parsed),
                inferred_schema=infer_output_schema(parsed),
                confidence_score=parsed.debug_info.confidence,
                extra_info=observed.extra_info,
            ),
            is_known_temp_table=is_known_temp_table,
            require_out_table_schema=require_out_table_schema,
            session_has_temp_tables=session_has_temp_tables,
            _is_internal=True,
        )

    def add_preparsed_query(
        self,
        parsed: PreparsedQuery,
        is_known_temp_table: bool = False,
        require_out_table_schema: bool = False,
        session_has_temp_tables: bool = True,
        _is_internal: bool = False,
    ) -> None:
        # Adding tool specific metadata extraction here allows it
        # to work for both ObservedQuery and PreparsedQuery as
        # add_preparsed_query it used within add_observed_query.
        self._tool_meta_extractor.extract_bi_metadata(parsed)

        if not _is_internal:
            self.report.num_preparsed_queries += 1

        if parsed.timestamp:
            # Sanity check - some of our usage subroutines require the timestamp to be in UTC.
            # Ideally we'd actually reject missing tzinfo too, but we can tighten that later.
            assert parsed.timestamp.tzinfo in {None, timezone.utc}

        query_fingerprint = parsed.query_id
        if not query_fingerprint:
            query_fingerprint = get_query_fingerprint(
                parsed.query_text,
                platform=self.platform.platform_name,
            )

        # Format the query.
        formatted_query = self._maybe_format_query(parsed.query_text)

        # Register the query's usage.
        if not self._usage_aggregator:
            pass  # usage is not enabled
        elif parsed.timestamp is None:
            self.report.usage_skipped_missing_timestamp += 1
        else:
            upstream_fields = parsed.column_usage or {}
            for upstream_urn in parsed.upstreams:
                # If the upstream table is a temp table or otherwise denied by filters, don't log usage for it.
                if not self.is_allowed_table(upstream_urn, allow_external=False) or (
                    require_out_table_schema
                    and not self._schema_resolver.has_urn(upstream_urn)
                ):
                    continue

                self._usage_aggregator.aggregate_event(
                    resource=upstream_urn,
                    start_time=parsed.timestamp,
                    query=formatted_query,
                    user=parsed.user.urn() if parsed.user else None,
                    fields=sorted(upstream_fields.get(upstream_urn, [])),
                    count=parsed.query_count,
                )

        if self._query_usage_counts is not None and parsed.timestamp is not None:
            assert self.usage_config is not None
            bucket = get_time_bucket(
                parsed.timestamp, self.usage_config.bucket_duration
            )
            counts = self._query_usage_counts.for_mutation(query_fingerprint, {})
            counts[bucket] = counts.get(bucket, 0) + parsed.query_count

        # Register the query.
        self._add_to_query_map(
            QueryMetadata(
                query_id=query_fingerprint,
                formatted_query_string=formatted_query,
                session_id=parsed.session_id,
                query_type=parsed.query_type,
                lineage_type=models.DatasetLineageTypeClass.TRANSFORMED,
                latest_timestamp=parsed.timestamp,
                actor=parsed.user,
                upstreams=parsed.upstreams,
                column_lineage=parsed.column_lineage or [],
                column_usage=parsed.column_usage or {},
                confidence_score=parsed.confidence_score,
                used_temp_tables=session_has_temp_tables,
                origin=parsed.origin,
            )
        )

        if not parsed.downstream:
            return
        out_table = parsed.downstream

        # Register the query's lineage.
        if (
            is_known_temp_table
            or (
                parsed.query_type.is_create()
                and parsed.query_type_props.get("temporary")
            )
            or self.is_temp_table(out_table)
            or (
                require_out_table_schema
                and not self._schema_resolver.has_urn(out_table)
            )
        ):
            # Infer the schema of the output table and track it for later.
            if parsed.inferred_schema is not None:
                self._inferred_temp_schemas[query_fingerprint] = parsed.inferred_schema

            # Also track the lineage for the temp table, for merging purposes later.
            self._temp_lineage_map.for_mutation(parsed.session_id, {}).setdefault(
                out_table, OrderedSet()
            ).add(query_fingerprint)

            # Also update schema resolver for missing session id
            if parsed.session_id == _MISSING_SESSION_ID and parsed.inferred_schema:
                self._missing_session_schema_resolver.add_temp_tables(
                    {out_table: parsed.inferred_schema}
                )

        else:
            # Non-temp tables immediately generate lineage.
            self._lineage_map.for_mutation(out_table, OrderedSet()).add(
                query_fingerprint
            )

    def add_table_rename(
        self,
        table_rename: TableRename,
    ) -> None:
        """Add a table rename to the aggregator.

        This will make all observed queries that reference the original urn
        will instead generate lineage for the new urn.
        """

        self.report.num_table_renames += 1

        # This will not work if the table is renamed multiple times.
        self._table_renames[table_rename.original_urn] = table_rename.new_urn

        original_table = self._name_from_urn(table_rename.original_urn)
        new_table = self._name_from_urn(table_rename.new_urn)

        self.add_preparsed_query(
            PreparsedQuery(
                query_id=None,
                query_text=table_rename.query
                or f"--Datahub generated query text--\n"
                f"alter table {original_table} rename to {new_table}",
                upstreams=[table_rename.original_urn],
                downstream=table_rename.new_urn,
                column_lineage=self._generate_identity_column_lineage(
                    downstream_urn=table_rename.new_urn,
                    upstream_urn=table_rename.original_urn,
                ),
                session_id=table_rename.session_id,
                timestamp=table_rename.timestamp,
            )
        )

    def add_table_swap(self, table_swap: TableSwap) -> None:
        """Add a table swap to the aggregator.

        Args:
            table_swap.urn1, table_swap.urn2: The dataset URNs to swap.
        """

        if table_swap.id() in self._table_swaps:
            # We have already processed this table swap once
            return

        self.report.num_table_swaps += 1
        self._table_swaps[table_swap.id()] = table_swap
        table1 = self._name_from_urn(table_swap.urn1)
        table2 = self._name_from_urn(table_swap.urn2)

        # NOTE: Both queries are different on purpose. Currently, we can not
        # store (A->B) and (B->A) lineage against same query.

        # NOTE: we do not store upstreams for temp table on purpose, as that would
        # otherwise overwrite original upstream query of temp table because
        # currently a temporay table can have only one upstream query.

        if not self.is_temp_table(table_swap.urn2):
            self.add_preparsed_query(
                PreparsedQuery(
                    query_id=None,
                    query_text=f"--Datahub generated query text--"
                    f"\nalter table {table1} swap with {table2}",
                    upstreams=[table_swap.urn1],
                    downstream=table_swap.urn2,
                    column_lineage=self._generate_identity_column_lineage(
                        upstream_urn=table_swap.urn1, downstream_urn=table_swap.urn2
                    ),
                    session_id=table_swap.session_id,
                    timestamp=table_swap.timestamp,
                )
            )

        if not self.is_temp_table(table_swap.urn1):
            self.add_preparsed_query(
                PreparsedQuery(
                    query_id=None,
                    query_text=f"--Datahub generated query text--\n"
                    f"alter table {table2} swap with {table1}",
                    upstreams=[table_swap.urn2],
                    downstream=table_swap.urn1,
                    column_lineage=self._generate_identity_column_lineage(
                        upstream_urn=table_swap.urn2, downstream_urn=table_swap.urn1
                    ),
                    session_id=table_swap.session_id,
                    timestamp=table_swap.timestamp,
                )
            )

    def _make_schema_resolver_for_session(
        self, session_id: str
    ) -> SchemaResolverInterface:
        schema_resolver: SchemaResolverInterface = self._schema_resolver
        if session_id == _MISSING_SESSION_ID:
            schema_resolver = self._missing_session_schema_resolver
        elif session_id in self._temp_lineage_map:
            temp_table_schemas: Dict[str, Optional[List[models.SchemaFieldClass]]] = {}
            for temp_table_urn, query_ids in self._temp_lineage_map[session_id].items():
                for query_id in query_ids:
                    temp_table_schemas[temp_table_urn] = (
                        self._inferred_temp_schemas.get(query_id)
                    )
                    if temp_table_schemas:
                        break

            if temp_table_schemas:
                schema_resolver = self._schema_resolver.with_temp_tables(
                    temp_table_schemas
                )
                self.report.num_queries_with_temp_tables_in_session += 1

        return schema_resolver

    def _process_view_definition(
        self, view_urn: UrnStr, view_definition: ViewDefinition
    ) -> None:
        # Note that in some cases, the view definition will be a SELECT statement
        # instead of a CREATE VIEW ... AS SELECT statement. In those cases, we can't
        # trust the parsed query type or downstream urn.

        # Run the SQL parser.
        parsed = self._run_sql_parser(
            view_definition.view_definition,
            default_db=view_definition.default_db,
            default_schema=view_definition.default_schema,
            schema_resolver=self._schema_resolver,
        )
        if parsed.debug_info.error:
            self.report.views_parse_failures[view_urn] = (
                f"{parsed.debug_info.error} on query: {view_definition.view_definition[:100]}"
            )
        if parsed.debug_info.table_error:
            self.report.num_views_failed += 1
            return  # we can't do anything with this query
        elif isinstance(parsed.debug_info.column_error, CooperativeTimeoutError):
            self.report.num_views_column_timeout += 1
        elif parsed.debug_info.column_error:
            self.report.num_views_column_failed += 1

        query_fingerprint = self._view_query_id(view_urn)
        formatted_view_definition = self._maybe_format_query(
            view_definition.view_definition
        )

        # Register the query.
        self._add_to_query_map(
            QueryMetadata(
                query_id=query_fingerprint,
                formatted_query_string=formatted_view_definition,
                session_id=_MISSING_SESSION_ID,
                query_type=QueryType.CREATE_VIEW,
                lineage_type=models.DatasetLineageTypeClass.VIEW,
                latest_timestamp=None,
                actor=None,
                upstreams=parsed.in_tables,
                column_lineage=parsed.column_lineage or [],
                column_usage=compute_upstream_fields(parsed),
                confidence_score=parsed.debug_info.confidence,
            )
        )

        # Register the query's lineage.
        self._lineage_map.for_mutation(view_urn, OrderedSet()).add(query_fingerprint)

    def _run_sql_parser(
        self,
        query: str,
        default_db: Optional[str],
        default_schema: Optional[str],
        schema_resolver: SchemaResolverInterface,
        session_id: str = _MISSING_SESSION_ID,
        timestamp: Optional[datetime] = None,
        user: Optional[Union[CorpUserUrn, CorpGroupUrn]] = None,
    ) -> SqlParsingResult:
        with self.report.sql_parsing_timer:
            parsed = sqlglot_lineage(
                query,
                schema_resolver=schema_resolver,
                default_db=default_db,
                default_schema=default_schema,
            )
        self.report.num_sql_parsed += 1

        # Conditionally log the query.
        if self.query_log == QueryLogSetting.STORE_ALL or (
            self.query_log == QueryLogSetting.STORE_FAILED and parsed.debug_info.error
        ):
            query_log_entry = LoggedQuery(
                query=query,
                session_id=session_id if session_id != _MISSING_SESSION_ID else None,
                timestamp=timestamp,
                user=user.urn() if user else None,
                default_db=default_db,
                default_schema=default_schema,
            )
            self._logged_queries.append(query_log_entry)

        # Also add some extra logging.
        if parsed.debug_info.error:
            logger.debug(
                f"Error parsing query {query}: {parsed.debug_info.error}",
                exc_info=parsed.debug_info.error,
            )

        return parsed

    def _add_to_query_map(
        self, new: QueryMetadata, merge_lineage: bool = False
    ) -> None:
        query_fingerprint = new.query_id

        if query_fingerprint in self._query_map:
            current = self._query_map.for_mutation(query_fingerprint)

            # This assumes that queries come in order of increasing timestamps,
            # so the current query is more authoritative than the previous one.
            current.formatted_query_string = new.formatted_query_string
            current.latest_timestamp = new.latest_timestamp or current.latest_timestamp
            current.actor = new.actor or current.actor

            if current.used_temp_tables and not new.used_temp_tables:
                # If we see the same query again, but in a different session,
                # it's possible that we didn't capture the temp tables in the newer session,
                # but did in the older one. If that happens, we treat the older session's
                # lineage as more authoritative. This isn't technically correct, but it's
                # better than using the newer session's lineage, which is likely incorrect.
                self.report.queries_with_non_authoritative_session.append(
                    query_fingerprint
                )
                return
            current.session_id = new.session_id

            if not merge_lineage:
                # An invariant of the fingerprinting is that if two queries have the
                # same fingerprint, they must also have the same lineage. We overwrite
                # here just in case more schemas got registered in the interim.
                current.upstreams = new.upstreams
                current.column_lineage = new.column_lineage
                current.column_usage = new.column_usage
                current.confidence_score = new.confidence_score
            else:
                # In the case of known query lineage, we might get things one at a time.
                # TODO: We don't yet support merging CLL for a single query.
                current.upstreams = list(
                    OrderedSet(current.upstreams) | OrderedSet(new.upstreams)
                )
                current.confidence_score = min(
                    current.confidence_score, new.confidence_score
                )
        else:
            self._query_map[query_fingerprint] = new

    def gen_metadata(self) -> Iterable[MetadataChangeProposalWrapper]:
        queries_generated: Set[QueryId] = set()

        yield from self._gen_lineage_mcps(queries_generated)
        self.report.num_queries_used_in_lineage = len(queries_generated)
        yield from self._gen_usage_statistics_mcps()
        yield from self._gen_operation_mcps(queries_generated)
        yield from self._gen_remaining_queries(queries_generated)

    def _gen_lineage_mcps(
        self, queries_generated: Set[QueryId]
    ) -> Iterable[MetadataChangeProposalWrapper]:
        if not self.generate_lineage:
            return

        # Process all views and inject them into the lineage map.
        # The parsing of view definitions is deferred until this point
        # to ensure the availability of all schema metadata.
        for view_urn, view_definition in self._view_definitions.items():
            self._process_view_definition(view_urn, view_definition)
        self._view_definitions.clear()

        # Generate lineage and queries.
        for downstream_urn in sorted(self._lineage_map):
            yield from self._gen_lineage_for_downstream(
                downstream_urn, queries_generated=queries_generated
            )

    @classmethod
    def _query_type_precedence(cls, query_type: str) -> int:
        query_precedence = [
            models.DatasetLineageTypeClass.COPY,
            models.DatasetLineageTypeClass.VIEW,
            models.DatasetLineageTypeClass.TRANSFORMED,
        ]

        # Lower value = higher precedence.
        idx = query_precedence.index(query_type)
        if idx == -1:
            return len(query_precedence)
        return idx

    def _gen_lineage_for_downstream(
        self, downstream_urn: str, queries_generated: Set[QueryId]
    ) -> Iterable[MetadataChangeProposalWrapper]:
        if not self.is_allowed_table(downstream_urn):
            self.report.num_lineage_skipped_due_to_filters += 1
            return

        query_ids = self._lineage_map[downstream_urn]
        queries: List[QueryMetadata] = [
            self._resolve_query_with_temp_tables(self._query_map[query_id])
            for query_id in query_ids
        ]

        # Sort the queries by highest precedence first, then by latest timestamp.
        # In case of ties, prefer queries with a known query type.
        # Tricky: by converting the timestamp to a number, we also can ignore the
        # differences between naive and aware datetimes.
        queries = sorted(
            # Sorted is a stable sort, so in the case of total ties, we want
            # to prefer the most recently added query.
            reversed(queries),
            key=lambda query: (
                self._query_type_precedence(query.lineage_type),
                -(make_ts_millis(query.latest_timestamp) or 0),
                query.query_type == QueryType.UNKNOWN,
            ),
        )

        queries_map: Dict[QueryId, QueryMetadata] = {
            query.query_id: query for query in queries
        }

        # mapping of upstream urn -> query id that produced it
        upstreams: Dict[UrnStr, QueryId] = {}
        # mapping of downstream column -> { upstream column -> query id that produced it }
        cll: Dict[str, Dict[SchemaFieldUrn, QueryId]] = defaultdict(dict)

        for query in queries:
            # Using setdefault to respect the precedence of queries.

            for upstream in query.upstreams:
                upstreams.setdefault(upstream, query.query_id)

            for lineage_info in query.column_lineage:
                for upstream_ref in lineage_info.upstreams:
                    cll[lineage_info.downstream.column].setdefault(
                        SchemaFieldUrn(upstream_ref.table, upstream_ref.column),
                        query.query_id,
                    )

        # Finally, we can build our lineage edge.
        required_queries = OrderedSet[QueryId]()
        upstream_aspect = models.UpstreamLineageClass(upstreams=[])
        for upstream_urn, query_id in upstreams.items():
            required_queries.add(query_id)

            upstream_aspect.upstreams.append(
                models.UpstreamClass(
                    dataset=upstream_urn,
                    type=queries_map[query_id].lineage_type,
                    query=(
                        self._query_urn(query_id)
                        if self.can_generate_query(query_id)
                        else None
                    ),
                    created=query.make_created_audit_stamp(),
                    auditStamp=models.AuditStampClass(
                        time=get_sys_time(),
                        actor=_DEFAULT_USER_URN.urn(),
                    ),
                )
            )
        upstream_aspect.fineGrainedLineages = []
        for downstream_column, all_upstream_columns in cll.items():
            # Group by query ID.
            for query_id, upstream_columns_for_query in groupby_unsorted(
                all_upstream_columns.items(),
                key=lambda x: x[1],
            ):
                upstream_columns = [x[0] for x in upstream_columns_for_query]
                required_queries.add(query_id)

                upstream_aspect.fineGrainedLineages.append(
                    models.FineGrainedLineageClass(
                        upstreamType=models.FineGrainedLineageUpstreamTypeClass.FIELD_SET,
                        upstreams=[
                            upstream_column.urn()
                            for upstream_column in upstream_columns
                        ],
                        downstreamType=models.FineGrainedLineageDownstreamTypeClass.FIELD,
                        downstreams=[
                            SchemaFieldUrn(downstream_urn, downstream_column).urn()
                        ],
                        query=(
                            self._query_urn(query_id)
                            if self.can_generate_query(query_id)
                            else None
                        ),
                        confidenceScore=queries_map[query_id].confidence_score,
                    )
                )

        if len(upstream_aspect.upstreams) > MAX_UPSTREAM_TABLES_COUNT:
            logger.warning(
                f"Too many upstream tables for {downstream_urn}: {len(upstream_aspect.upstreams)}"
                f"Keeping only {MAX_UPSTREAM_TABLES_COUNT} table level upstreams/"
            )
            upstream_aspect.upstreams = upstream_aspect.upstreams[
                :MAX_UPSTREAM_TABLES_COUNT
            ]
            self.report.num_table_lineage_trimmed_due_to_large_size += 1
        if len(upstream_aspect.fineGrainedLineages) > MAX_FINEGRAINEDLINEAGE_COUNT:
            logger.warning(
                f"Too many upstream columns for {downstream_urn}: {len(upstream_aspect.fineGrainedLineages)}"
                f"Keeping only {MAX_FINEGRAINEDLINEAGE_COUNT} column level upstreams/"
            )
            upstream_aspect.fineGrainedLineages = upstream_aspect.fineGrainedLineages[
                :MAX_FINEGRAINEDLINEAGE_COUNT
            ]
            self.report.num_column_lineage_trimmed_due_to_large_size += 1

        upstream_aspect.fineGrainedLineages = (
            upstream_aspect.fineGrainedLineages or None
        )

        if not upstream_aspect.upstreams and not upstream_aspect.fineGrainedLineages:
            return

        yield MetadataChangeProposalWrapper(
            entityUrn=downstream_urn,
            aspect=upstream_aspect,
        )

        for query_id in required_queries:
            # Avoid generating the same query twice.
            if query_id in queries_generated:
                continue
            queries_generated.add(query_id)

            query = queries_map[query_id]
            yield from self._gen_query(query, downstream_urn)

    @classmethod
    def _query_urn(cls, query_id: QueryId) -> str:
        return QueryUrn(query_id).urn()

    @classmethod
    def _composite_query_id(cls, composed_of_queries: List[QueryId]) -> str:
        combined = json.dumps(composed_of_queries)
        return f"composite_{generate_hash(combined)}"

    @classmethod
    def _view_query_id(cls, view_urn: UrnStr) -> str:
        return f"view_{DatasetUrn.url_encode(view_urn)}"

    @classmethod
    def _known_lineage_query_id(cls) -> str:
        return f"known_{uuid.uuid4()}"

    @classmethod
    def _is_known_lineage_query_id(cls, query_id: QueryId) -> bool:
        # Our query fingerprints are hex and won't have underscores, so this will
        # never conflict with a real query fingerprint.
        return query_id.startswith("known_")

    def _gen_remaining_queries(
        self, queries_generated: Set[QueryId]
    ) -> Iterable[MetadataChangeProposalWrapper]:
        if not self.generate_queries or not self.generate_query_usage_statistics:
            return

        assert self._query_usage_counts is not None
        for query_id in self._query_usage_counts:
            if query_id in queries_generated:
                continue
            queries_generated.add(query_id)

            yield from self._gen_query(self._query_map[query_id])

    def can_generate_query(self, query_id: QueryId) -> bool:
        return self.generate_queries and not self._is_known_lineage_query_id(query_id)

    def _gen_query(
        self, query: QueryMetadata, downstream_urn: Optional[str] = None
    ) -> Iterable[MetadataChangeProposalWrapper]:
        query_id = query.query_id
        if not self.can_generate_query(query_id):
            return

        # If a query doesn't involve any allowed tables, skip it.
        if downstream_urn is None and not any(
            self.is_allowed_table(urn) for urn in query.upstreams
        ):
            self.report.num_queries_skipped_due_to_filters += 1
            return

        yield from MetadataChangeProposalWrapper.construct_many(
            entityUrn=self._query_urn(query_id),
            aspects=[
                query.make_query_properties(),
                make_query_subjects(
                    query.get_subjects(
                        downstream_urn=downstream_urn,
                        include_fields=self.generate_query_subject_fields,
                    )
                ),
                models.DataPlatformInstanceClass(
                    platform=self.platform.urn(),
                ),
            ],
        )
        self.report.num_queries_entities_generated += 1

        if self._query_usage_counts is not None:
            assert self.usage_config is not None

            # This is slightly lossy, since we only store one unique
            # user per query instead of tracking all of them.
            # We also lose information because we don't keep track
            # of users / lastExecutedAt timestamps per bucket.
            user = query.actor

            query_counter = self._query_usage_counts.get(query_id)
            if not query_counter:
                return

            all_buckets = self.usage_config.buckets()

            for bucket, count in query_counter.items():
                if bucket not in all_buckets:
                    # What happens if we get a query with a timestamp that's outside our configured window?
                    # Theoretically this should never happen, since the audit logs are also fetched
                    # for the window. However, it's useful to have reporting for it, just in case.
                    self.report.num_query_usage_stats_outside_window += 1
                    continue

                yield MetadataChangeProposalWrapper(
                    entityUrn=self._query_urn(query_id),
                    aspect=models.QueryUsageStatisticsClass(
                        timestampMillis=make_ts_millis(bucket),
                        eventGranularity=models.TimeWindowSizeClass(
                            unit=self.usage_config.bucket_duration, multiple=1
                        ),
                        queryCount=count,
                        uniqueUserCount=1,
                        userCounts=(
                            [
                                models.DatasetUserUsageCountsClass(
                                    user=user.urn(),
                                    count=count,
                                )
                            ]
                            if user
                            else None
                        ),
                    ),
                )

                self.report.num_query_usage_stats_generated += 1

    def _resolve_query_with_temp_tables(
        self,
        base_query: QueryMetadata,
    ) -> QueryMetadata:
        # TODO: Add special handling for COPY operations, which should mirror the schema
        # of the thing being copied in order to generate CLL.

        session_id = base_query.session_id

        composed_of_queries = OrderedSet[QueryId]()

        @dataclasses.dataclass
        class QueryLineageInfo:
            upstreams: List[UrnStr]  # this is direct upstreams, with *no temp tables*
            column_lineage: List[ColumnLineageInfo]
            confidence_score: float

            def _merge_lineage_from(self, other_query: "QueryLineageInfo") -> None:
                self.upstreams += other_query.upstreams
                self.column_lineage += other_query.column_lineage
                self.confidence_score = min(
                    self.confidence_score, other_query.confidence_score
                )

        def _recurse_into_query(
            query: QueryMetadata, recursion_path: List[QueryId]
        ) -> QueryLineageInfo:
            if query.query_id in recursion_path:
                # This is a cycle, so we just return the query as-is.
                return QueryLineageInfo(
                    upstreams=query.upstreams,
                    column_lineage=query.column_lineage,
                    confidence_score=query.confidence_score,
                )
            recursion_path = [*recursion_path, query.query_id]
            composed_of_queries.add(query.query_id)

            # Find all the temp tables that this query depends on.
            temp_upstream_queries: Dict[UrnStr, QueryLineageInfo] = {}
            for upstream in query.upstreams:
                upstream_query_ids = self._temp_lineage_map.get(session_id, {}).get(
                    upstream
                )
                if upstream_query_ids:
                    for upstream_query_id in upstream_query_ids:
                        upstream_query = self._query_map.get(upstream_query_id)
                        if (
                            upstream_query
                            and upstream_query.query_id not in composed_of_queries
                        ):
                            temp_query_lineage_info = _recurse_into_query(
                                upstream_query, recursion_path
                            )
                            if upstream in temp_upstream_queries:
                                temp_upstream_queries[upstream]._merge_lineage_from(
                                    temp_query_lineage_info
                                )
                            else:
                                temp_upstream_queries[upstream] = (
                                    temp_query_lineage_info
                                )

            # Compute merged upstreams.
            new_upstreams = OrderedSet[UrnStr]()
            for upstream in query.upstreams:
                if upstream in temp_upstream_queries:
                    new_upstreams.update(temp_upstream_queries[upstream].upstreams)
                else:
                    new_upstreams.add(upstream)

            # Compute merged column lineage.
            new_cll = []
            for lineage_info in query.column_lineage:
                new_column_upstreams: List[ColumnRef] = []
                for existing_col_upstream in lineage_info.upstreams:
                    if existing_col_upstream.table in temp_upstream_queries:
                        new_column_upstreams.extend(
                            [
                                temp_col_upstream
                                for temp_lineage_info in temp_upstream_queries[
                                    existing_col_upstream.table
                                ].column_lineage
                                for temp_col_upstream in temp_lineage_info.upstreams
                                if temp_lineage_info.downstream.column
                                == existing_col_upstream.column
                            ]
                        )
                    else:
                        new_column_upstreams.append(existing_col_upstream)

                new_cll.append(
                    ColumnLineageInfo(
                        downstream=lineage_info.downstream,
                        upstreams=new_column_upstreams,
                    )
                )

            # Compute merged confidence score.
            new_confidence_score = min(
                [
                    query.confidence_score,
                    *[
                        temp_upstream_query.confidence_score
                        for temp_upstream_query in temp_upstream_queries.values()
                    ],
                ]
            )

            return QueryLineageInfo(
                upstreams=list(new_upstreams),
                column_lineage=new_cll,
                confidence_score=new_confidence_score,
            )

        resolved_lineage_info = _recurse_into_query(base_query, [])

        # Fast path if there were no temp tables.
        if len(composed_of_queries) == 1:
            return base_query

        # If this query does actually depend on temp tables:
        # - Clone the query into a new object
        # - Generate a new composite fingerprint
        # - Update the lineage info
        # - Update the query text to combine the queries

        ordered_queries = [
            self._query_map[query_id] for query_id in reversed(composed_of_queries)
        ]
        if all(q.latest_timestamp for q in ordered_queries):
            ordered_queries = sorted(
                ordered_queries,
                key=lambda query: make_ts_millis(query.latest_timestamp) or 0,
            )
        composite_query_id = self._composite_query_id(
            [q.query_id for q in ordered_queries]
        )
        composed_of_queries_truncated: LossyList[str] = LossyList()
        for query_id in composed_of_queries:
            composed_of_queries_truncated.append(query_id)
        self.report.queries_with_temp_upstreams[composite_query_id] = (
            composed_of_queries_truncated
        )

        merged_query_text = ";\n\n".join(
            [q.formatted_query_string for q in ordered_queries]
        )

        resolved_query = dataclasses.replace(
            base_query,
            query_id=composite_query_id,
            formatted_query_string=merged_query_text,
            upstreams=resolved_lineage_info.upstreams,
            column_lineage=resolved_lineage_info.column_lineage,
            confidence_score=resolved_lineage_info.confidence_score,
        )

        return resolved_query

    def _gen_usage_statistics_mcps(self) -> Iterable[MetadataChangeProposalWrapper]:
        if not self._usage_aggregator:
            return

        for wu in self._usage_aggregator.generate_workunits(
            resource_urn_builder=lambda urn: urn,
            user_urn_builder=lambda urn: urn,
        ):
            # TODO: We should change the usage aggregator to return MCPWs directly.
            yield cast(MetadataChangeProposalWrapper, wu.metadata)

    def _gen_operation_mcps(
        self, queries_generated: Set[QueryId]
    ) -> Iterable[MetadataChangeProposalWrapper]:
        if not self.generate_operations:
            return

        for downstream_urn, query_ids in self._lineage_map.items():
            for query_id in query_ids:
                yield from self._gen_operation_for_downstream(downstream_urn, query_id)

                # Avoid generating the same query twice.
                if query_id in queries_generated:
                    continue
                queries_generated.add(query_id)
                yield from self._gen_query(self._query_map[query_id], downstream_urn)

    def _gen_operation_for_downstream(
        self, downstream_urn: UrnStr, query_id: QueryId
    ) -> Iterable[MetadataChangeProposalWrapper]:
        query = self._query_map[query_id]
        if query.latest_timestamp is None:
            return

        operation_type = query.query_type.to_operation_type()
        if operation_type is None:
            # We don't generate operations for SELECTs.
            return

        if not self.is_allowed_table(downstream_urn):
            self.report.num_operations_skipped_due_to_filters += 1
            return

        self.report.num_operations_generated += 1
        aspect = models.OperationClass(
            timestampMillis=make_ts_millis(datetime.now(tz=timezone.utc)),
            operationType=operation_type,
            lastUpdatedTimestamp=make_ts_millis(query.latest_timestamp),
            actor=query.actor.urn() if query.actor else None,
            customProperties=(
                {"query_urn": self._query_urn(query_id)}
                if self.can_generate_query(query_id)
                else None
            ),
        )
        yield MetadataChangeProposalWrapper(entityUrn=downstream_urn, aspect=aspect)



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Union, overload

import datahub.metadata.schema_classes as models
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.emitter.mcp_patch_builder import MetadataPatchProposal
from datahub.errors import IngestionAttributionWarning, ItemNotFoundError, SdkUsageError
from datahub.ingestion.graph.client import DataHubGraph
from datahub.metadata.urns import (
    ContainerUrn,
    DatasetUrn,
    MlModelGroupUrn,
    MlModelUrn,
    Urn,
)
from datahub.sdk._all_entities import ENTITY_CLASSES
from datahub.sdk._shared import UrnOrStr
from datahub.sdk.container import Container
from datahub.sdk.dataset import Dataset
from datahub.sdk.entity import Entity
from datahub.sdk.mlmodel import MLModel
from datahub.sdk.mlmodelgroup import MLModelGroup

if TYPE_CHECKING:
    from datahub.sdk.main_client import DataHubClient


class EntityClient:
    """Client for managing DataHub entities.

    This class provides methods for retrieving and managing DataHub entities
    such as datasets, containers, and other metadata objects.
    """

    def __init__(self, client: DataHubClient):
        """Private constructor - use :py:attr:`DataHubClient.entities` instead.

        Args:
            client: The parent DataHubClient instance.
        """
        self._client = client

    # TODO: Make all of these methods sync by default.

    @property
    def _graph(self) -> DataHubGraph:
        return self._client._graph

    @overload
    def get(self, urn: ContainerUrn) -> Container: ...
    @overload
    def get(self, urn: DatasetUrn) -> Dataset: ...
    @overload
    def get(self, urn: MlModelUrn) -> MLModel: ...
    @overload
    def get(self, urn: MlModelGroupUrn) -> MLModelGroup: ...
    @overload
    def get(self, urn: Union[Urn, str]) -> Entity: ...
    def get(self, urn: UrnOrStr) -> Entity:
        """Retrieve an entity by its urn.

        Args:
            urn: The urn of the entity to retrieve. Can be a string or :py:class:`Urn` object.

        Returns:
            The retrieved entity instance.

        Raises:
            ItemNotFoundError: If the entity does not exist.
            SdkUsageError: If the entity type is not yet supported.
            InvalidUrnError: If the URN is invalid.
        """
        if not isinstance(urn, Urn):
            urn = Urn.from_string(urn)

        # TODO: add error handling around this with a suggested alternative if not yet supported
        EntityClass = ENTITY_CLASSES[urn.entity_type]

        if not self._graph.exists(str(urn)):
            raise ItemNotFoundError(f"Entity {urn} not found")

        aspects = self._graph.get_entity_semityped(str(urn))

        # TODO: save the timestamp so we can use If-Unmodified-Since on the updates
        return EntityClass._new_from_graph(urn, aspects)

    def create(self, entity: Entity) -> None:
        mcps = []

        if self._graph.exists(str(entity.urn)):
            raise SdkUsageError(
                f"Entity {entity.urn} already exists. Use client.entities.upsert() to update it."
            )

        # Extra safety check: by putting this first, we can ensure that
        # the request fails if the entity already exists.
        mcps.append(
            MetadataChangeProposalWrapper(
                entityUrn=str(entity.urn),
                aspect=entity.urn.to_key_aspect(),
                changeType=models.ChangeTypeClass.CREATE_ENTITY,
            )
        )
        mcps.extend(entity.as_mcps(models.ChangeTypeClass.CREATE))

        self._graph.emit_mcps(mcps)

    def upsert(self, entity: Entity) -> None:
        if entity._prev_aspects is None and self._graph.exists(str(entity.urn)):
            warnings.warn(
                f"The entity {entity.urn} already exists. This operation will partially overwrite the existing entity.",
                IngestionAttributionWarning,
                stacklevel=2,
            )
            # TODO: If there are no previous aspects but the entity exists, should we delete aspects that are not present here?

        mcps = entity.as_mcps(models.ChangeTypeClass.UPSERT)
        self._graph.emit_mcps(mcps)

    def update(self, entity: Union[Entity, MetadataPatchProposal]) -> None:
        if isinstance(entity, MetadataPatchProposal):
            return self._update_patch(entity)

        if entity._prev_aspects is None:
            raise SdkUsageError(
                f"For entities created via {entity.__class__.__name__}(...), use client.entities.create() or client.entities.upsert() instead"
            )

        # TODO: respect If-Unmodified-Since?
        # -> probably add a "mode" parameter that can be "update" (e.g. if not modified) or "update_force"

        mcps = entity.as_mcps(models.ChangeTypeClass.UPSERT)
        self._graph.emit_mcps(mcps)

    def _update_patch(
        self, updater: MetadataPatchProposal, check_exists: bool = True
    ) -> None:
        if check_exists and not self._graph.exists(updater.urn):
            raise SdkUsageError(
                f"Entity {updater.urn} does not exist, and hence cannot be updated. "
                "You can bypass this check by setting check_exists=False."
            )

        mcps = updater.build()
        self._graph.emit_mcps(mcps)



import json
import logging
from dataclasses import dataclass
from typing import Any, Dict, Iterable, List, Optional, Tuple

from pydantic import BaseModel, Field, SecretStr
from slack_sdk import WebClient
from tenacity import retry, wait_exponential
from tenacity.before_sleep import before_sleep_log

import datahub.emitter.mce_builder as builder
from datahub.emitter.mce_builder import datahub_guid, make_dataplatform_instance_urn
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.decorators import (
    SupportStatus,
    config_class,
    platform_name,
    support_status,
)
from datahub.ingestion.api.source import (
    MetadataWorkUnitProcessor,
    SourceReport,
)
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.source.state.stale_entity_removal_handler import (
    StaleEntityRemovalHandler,
    StaleEntityRemovalSourceReport,
)
from datahub.ingestion.source.state.stateful_ingestion_base import (
    StatefulIngestionConfigBase,
    StatefulIngestionSourceBase,
)
from datahub.metadata.schema_classes import (
    CorpUserEditableInfoClass,
    CorpUserSettingsClass,
    DataPlatformInstanceClass,
    DataPlatformInstancePropertiesClass,
    DatasetPropertiesClass,
    DeprecationClass,
    NotificationSettingsClass,
    PlatformResourceInfoClass,
    SerializedValueClass,
    SerializedValueContentTypeClass,
    SerializedValueSchemaTypeClass,
    SlackNotificationSettingsClass,
    SlackUserInfoClass as SlackUserInfo,
    StatusClass,
    SubTypesClass,
    _Aspect,
)
from datahub.utilities.ratelimiter import RateLimiter
from datahub.utilities.str_enum import StrEnum
from datahub.utilities.urns.urn import Urn

logger: logging.Logger = logging.getLogger(__name__)


# TODO: Relocate this function to a utility module
def is_picture_default_or_missing(picture_link: Optional[str]) -> bool:
    if not picture_link:
        return True
    return picture_link.endswith("default_avatar.png")


def is_slack_image(picture_link: Optional[str]) -> bool:
    """
    Guesses if the picture link is a slack image.
    """
    if not picture_link:
        return False
    return "slack-edge.com" in picture_link


class ResourceType(StrEnum):
    USER_INFO = "user-info"
    CHANNEL_INFO = "channel-info"


class SlackInstance(BaseModel):
    id: str
    name: Optional[str] = None
    description: Optional[str] = None
    external_url: Optional[str] = None
    custom_properties: Optional[Dict[str, str]] = None

    def to_platform_instance_urn(self) -> str:
        return make_dataplatform_instance_urn(
            platform=DATA_PLATFORM_SLACK_URN, instance=self.id
        )

    def with_slack_team_info(self, team_info: dict) -> "SlackInstance":
        """
        team_info looks like this
        {'id': 'T22BUCL1LKW', 'name': 'DataHub', 'url': 'https://datahubspace.slack.com/', 'domain': 'datahub', 'email_domain': '', 'icon': {'image_default': False, 'image_34': 'https://avatars.slack-edge.com/2021-07-05/2228585180071_63e6f300a919abc488bb_34.png', 'image_44': 'https://avatars.slack-edge.com/2021-07-05/2228585180071_63e6f300a919abc488bb_44.png', 'image_68': 'https://avatars.slack-edge.com/2021-07-05/2228585180071_63e6f300a919abc488bb_68.png', 'image_88': 'https://avatars.slack-edge.com/2021-07-05/2228585180071_63e6f300a919abc488bb_88.png', 'image_102': 'https://avatars.slack-edge.com/2021-07-05/2228585180071_63e6f300a919abc488bb_102.png', 'image_230': 'https://avatars.slack-edge.com/2021-07-05/2228585180071_63e6f300a919abc488bb_230.png', 'image_132': 'https://avatars.slack-edge.com/2021-07-05/2228585180071_63e6f300a919abc488bb_132.png'}, 'avatar_base_url': 'https://ca.slack-edge.com/', 'is_verified': False, 'external_org_migrations': {'date_updated': 1722672564, 'current': []}, 'discoverable': 'closed', 'enterprise_id': 'E06TPM5T1G9', 'enterprise_name': 'DataHub', 'enterprise_domain': 'datahubspace', 'lob_sales_home_enabled': False}
        """
        self.name = team_info.get("name")
        self.description = team_info.get("name")
        self.external_url = team_info.get("url")
        self.custom_properties = {
            k: v
            for k, v in {
                "domain": team_info.get("domain"),
                "enterprise_id": team_info.get("enterprise_id"),
                "enterprise_name": team_info.get("enterprise_name"),
                "enterprise_domain": team_info.get("enterprise_domain"),
                "icon": team_info.get("icon", {}).get("image_102"),
            }.items()
            if v is not None
        }
        return self

    def to_mcps(self) -> Iterable[MetadataChangeProposalWrapper]:
        return [
            MetadataChangeProposalWrapper(
                entityUrn=self.to_platform_instance_urn(),
                aspect=DataPlatformInstancePropertiesClass(
                    name=self.name or self.id,
                    description=self.description,
                    externalUrl=self.external_url or None,
                    customProperties=self.custom_properties or {},
                ),
            )
        ]


def to_serialized_value(value: _Aspect) -> SerializedValueClass:
    # HACK: we remove the .pegasus2avro from the schema type since we want to refer to
    # the original pdl type
    schema_type = value.RECORD_SCHEMA.fullname.replace(".pegasus2avro", "")
    serialized_value = SerializedValueClass(
        blob=json.dumps(value.to_obj()).encode("utf-8"),
        contentType=SerializedValueContentTypeClass.JSON,
        schemaType=SerializedValueSchemaTypeClass.PEGASUS,
        schemaRef=schema_type,
    )
    return serialized_value


class SlackUserDetails:
    def __init__(self, slack_user_info: SlackUserInfo):
        self.slack_user_info = slack_user_info

    def to_guid(self) -> str:
        """
        A slack user is uniquely identified by the combination of their id and teamId.
        """
        return datahub_guid(
            {"id": self.slack_user_info.id, "dpi": self.slack_user_info.teamId}
        )

    def get_resource_urn(self) -> str:
        return f"urn:li:platformResource:{self.to_guid()}"

    def to_mcps(self) -> Iterable[MetadataChangeProposalWrapper]:
        resource_urn = self.get_resource_urn()

        dpi = DataPlatformInstanceClass(
            platform=DATA_PLATFORM_SLACK_URN,
            instance=self.slack_user_info.slackInstance,
        )

        secondary_keys = []
        if self.slack_user_info.email:
            secondary_keys.append(self.slack_user_info.email)

        resource_info = PlatformResourceInfoClass(
            resourceType=ResourceType.USER_INFO.value,
            value=to_serialized_value(self.slack_user_info),
            primaryKey=self.slack_user_info.id,
            secondaryKeys=secondary_keys,
        )

        status = StatusClass(
            removed=self.slack_user_info.isDeleted,
        )

        yield from MetadataChangeProposalWrapper.construct_many(
            resource_urn, aspects=[dpi, resource_info, status]
        )


@dataclass
class CorpUser:
    urn: Optional[str] = None
    email: Optional[str] = None
    slack_id: Optional[str] = None
    title: Optional[str] = None
    image_url: Optional[str] = None
    phone: Optional[str] = None
    real_name: Optional[str] = None
    slack_display_name: Optional[str] = None
    team_id: Optional[str] = None
    team_domain: Optional[str] = None
    is_team_enterprise: Optional[bool] = None


class SlackSourceConfig(
    StatefulIngestionConfigBase,
):
    bot_token: SecretStr = Field(
        description="Bot token for the Slack workspace. Needs `users:read`, `users:read.email`, `users.profile:read`, and `team:read` scopes.",
    )
    enrich_user_metadata: bool = Field(
        type=bool,
        default=True,
        description="When enabled, will enrich provisioned DataHub users' metadata with information from Slack.",
    )
    ingest_users: bool = Field(
        type=bool,
        default=True,
        description="Whether to ingest users. When set to true, will ingest all users in the Slack workspace (as platform resources) to simplify user enrichment after they are provisioned on DataHub.",
    )
    api_requests_per_min: int = Field(
        type=int,
        default=10,
        description="Number of API requests per minute. Low-level config. Do not tweak unless you are facing any issues.",
    )
    ingest_public_channels: bool = Field(
        type=bool,
        default=False,
        description="Whether to ingest public channels. If set to true needs `channels:read` scope.",
    )
    channels_iteration_limit: int = Field(
        type=int,
        default=200,
        description="Limit the number of channels to be ingested in a iteration. Low-level config. Do not tweak unless you are facing any issues.",
    )
    channel_min_members: int = Field(
        type=int,
        default=2,
        description="Ingest channels with at least this many members.",
    )
    should_ingest_archived_channels: bool = Field(
        type=bool,
        default=False,
        description="Whether to ingest archived channels.",
    )


@dataclass
class SlackSourceReport(StaleEntityRemovalSourceReport):
    channels_reported: int = 0
    archived_channels_reported: int = 0
    users_reported: int = 0


PLATFORM_NAME = "slack"
DATA_PLATFORM_SLACK_URN: str = builder.make_data_platform_urn(PLATFORM_NAME)


@platform_name("Slack")
@config_class(SlackSourceConfig)
@support_status(SupportStatus.TESTING)
class SlackSource(StatefulIngestionSourceBase):
    def __init__(self, ctx: PipelineContext, config: SlackSourceConfig):
        super().__init__(config, ctx)
        self.ctx = ctx
        self.config = config
        self.report: SlackSourceReport = SlackSourceReport()
        self.workspace_base_url: Optional[str] = None
        self.rate_limiter = RateLimiter(
            max_calls=self.config.api_requests_per_min, period=60
        )
        self._use_users_info = False

    @classmethod
    def create(cls, config_dict, ctx):
        config = SlackSourceConfig.parse_obj(config_dict)
        return cls(ctx, config)

    def get_slack_client(self) -> WebClient:
        return WebClient(token=self.config.bot_token.get_secret_value())

    @staticmethod
    def populate_slack_member_from_response(
        user: Dict[str, Any], slack_instance: SlackInstance
    ) -> SlackUserDetails:
        profile = user.get("profile", {})

        user_info = SlackUserInfo(
            slackInstance=slack_instance.to_platform_instance_urn(),
            id=user["id"],
            name=user["name"],
            realName=user.get("real_name", ""),
            displayName=profile.get("display_name", ""),
            email=profile.get("email"),
            teamId=user["team_id"],
            isDeleted=user.get("deleted", False),
            isAdmin=user.get("is_admin", False),
            isOwner=user.get("is_owner", False),
            isPrimaryOwner=user.get("is_primary_owner", False),
            isBot=user.get("is_bot", False),
            timezone=user.get("tz"),
            timezoneOffset=user.get("tz_offset"),
            title=profile.get("title"),
            phone=profile.get("phone"),
            profilePictureUrl=profile.get(
                "image_192"
            ),  # Using 192px image as an example
            statusText=profile.get("status_text"),
            statusEmoji=profile.get("status_emoji"),
            lastUpdatedSeconds=user.get("updated"),
        )
        return SlackUserDetails(slack_user_info=user_info)

    def get_workunit_processors(self) -> List[Optional[MetadataWorkUnitProcessor]]:
        return [
            *super().get_workunit_processors(),
            StaleEntityRemovalHandler.create(
                self, self.config, self.ctx
            ).workunit_processor,
        ]

    def get_workunits_internal(
        self,
    ) -> Iterable[MetadataWorkUnit]:
        assert self.ctx.graph is not None
        auth_resp = self.get_slack_client().auth_test()
        assert isinstance(auth_resp.data, dict)
        self.workspace_base_url = str(auth_resp.data.get("url"))
        logger.info("Successfully connected to Slack")
        logger.info(auth_resp.data)
        if self.config.ingest_public_channels:
            yield from self.get_public_channels()
        if self.config.enrich_user_metadata or self.config.ingest_users:
            yield from self.get_user_info()

    def _get_datahub_user_info(
        self,
    ) -> Dict[str, Tuple[CorpUser, Optional[CorpUserEditableInfoClass]]]:
        # get_user_to_be_updated ensures that the email field is not None
        users = {
            user_obj.email: (user_obj, editable_properties)
            for user_obj, editable_properties in self.get_user_to_be_updated()
            if user_obj.email
        }
        return users

    def get_user_info(self) -> Iterable[MetadataWorkUnit]:
        # Get team information to populate for all users
        slack_instance: Optional[SlackInstance] = None
        with self.rate_limiter:
            team_response = self.get_slack_client().team_info()
            if team_response and "team" in team_response:
                team_info = team_response["team"]
                slack_instance = SlackInstance(id=team_info.get("id"))
                slack_instance = slack_instance.with_slack_team_info(team_info)

        if slack_instance:
            for mcp in slack_instance.to_mcps():
                yield mcp.as_workunit()
        else:
            logger.error("Failed to fetch team information")
            self.report.report_failure(
                "team_info", "Failed to fetch team information for users"
            )

        assert slack_instance

        # Fetch all DataHub users that need to be updated
        if self.config.enrich_user_metadata:
            datahub_users = self._get_datahub_user_info()
        else:
            datahub_users = {}
        cursor = None
        while True:
            with self.rate_limiter:
                response = self.get_slack_client().users_list(cursor=cursor)
            assert isinstance(response.data, dict)
            if not response.data["ok"]:
                self.report.report_failure("users", "Failed to fetch users")
                return

            assert self.ctx.graph is not None
            for user in response.data["members"]:
                # Query all slack users and ingest them into the generic
                # slackMember aspect
                slack_user_details: SlackUserDetails = (
                    self.populate_slack_member_from_response(user, slack_instance)
                )
                if self.config.ingest_users:
                    for mcp in slack_user_details.to_mcps():
                        yield mcp.as_workunit()

                platform_resource_urn = slack_user_details.get_resource_urn()
                # If user is in DataHub, compute and emit CorpUserEditableInfo
                # aspect. This code will be removed once we have server side
                # processing of raw slackMember aspects. This code path can also
                # be turned off by setting enrich_user_metadata to False.
                user_obj_props_tuple = datahub_users.get(user["profile"].get("email"))
                if user_obj_props_tuple is None:
                    # User is not in DataHub or enrichment is disabled
                    continue
                user_obj, editable_properties = user_obj_props_tuple
                slack_user_profile = user.get("profile", {})
                user_obj.slack_id = user.get("id")
                user_obj.title = slack_user_profile.get("title")
                user_obj.image_url = slack_user_profile.get("image_192")
                user_obj.phone = slack_user_profile.get("phone")
                user_obj.real_name = slack_user_profile.get("real_name")
                user_obj.slack_display_name = slack_user_profile.get("display_name")
                corpuser_editable_info = editable_properties or (
                    CorpUserEditableInfoClass()
                )
                emittable_corpuser_editable_info = self.populate_corpuser_editable_info(
                    corpuser_editable_info,
                    user_obj,
                    platform_resource_urn=platform_resource_urn,
                    slack_instance=slack_instance,
                )
                if emittable_corpuser_editable_info:
                    yield MetadataChangeProposalWrapper(
                        entityUrn=user_obj.urn, aspect=emittable_corpuser_editable_info
                    ).as_workunit()
                    # if we update corpusereditable info, we also update
                    # slackuserinfo. This will be removed once we have server
                    # side processing of raw slackMember aspects.
                    yield MetadataChangeProposalWrapper(
                        entityUrn=user_obj.urn,
                        aspect=slack_user_details.slack_user_info,
                    ).as_workunit()
                yield from self.emit_corp_user_slack_settings(user_obj)
            cursor = str(response.data["response_metadata"]["next_cursor"])
            if not cursor:
                break

    def _get_channel_info(
        self, cursor: Optional[str]
    ) -> Tuple[List[MetadataWorkUnit], Optional[str]]:
        result_channels: List[MetadataWorkUnit] = []
        with self.rate_limiter:
            response = self.get_slack_client().conversations_list(
                types="public_channel",
                limit=self.config.channels_iteration_limit,
                cursor=cursor,
            )
        assert isinstance(response.data, dict)
        if not response.data["ok"]:
            self.report.report_failure(
                "public_channel", "Failed to fetch public channels"
            )
            return result_channels, None
        for channel in response.data["channels"]:
            num_members = channel["num_members"]
            if num_members < self.config.channel_min_members:
                continue
            channel_id = channel["id"]
            urn_channel = builder.make_dataset_urn(
                platform=PLATFORM_NAME, name=channel_id
            )
            name = channel["name"]
            is_archived = channel.get("is_archived", False)
            if is_archived:
                if not self.config.should_ingest_archived_channels:
                    continue
                self.report.archived_channels_reported += 1
                logger.info(f"Archived channel: {name}")
                result_channels.append(
                    MetadataWorkUnit(
                        id=f"{urn_channel}-deprecation",
                        mcp=MetadataChangeProposalWrapper(
                            entityUrn=urn_channel,
                            aspect=DeprecationClass(
                                deprecated=True,
                                note="This channel is archived",
                                actor="urn:li:corpuser:datahub",
                            ),
                        ),
                    )
                )

            topic = channel.get("topic", {}).get("value")
            purpose = channel.get("purpose", {}).get("value")
            if self.workspace_base_url:
                external_url = f"{self.workspace_base_url}/archives/{channel_id}"
            result_channels.append(
                MetadataWorkUnit(
                    id=f"{urn_channel}-datasetproperties",
                    mcp=MetadataChangeProposalWrapper(
                        entityUrn=urn_channel,
                        aspect=DatasetPropertiesClass(
                            name=name,
                            externalUrl=external_url,
                            description=f"Topic: {topic}\nPurpose: {purpose}",
                        ),
                    ),
                )
            )
            result_channels.append(
                MetadataWorkUnit(
                    id=f"{urn_channel}-subtype",
                    mcp=MetadataChangeProposalWrapper(
                        entityUrn=urn_channel,
                        aspect=SubTypesClass(
                            typeNames=["Slack Channel"],
                        ),
                    ),
                )
            )
        cursor = str(response.data["response_metadata"]["next_cursor"])
        return result_channels, cursor

    def populate_corpuser_editable_info(
        self,
        corpuser_editable_info: CorpUserEditableInfoClass,
        user_obj: CorpUser,
        platform_resource_urn: str,
        slack_instance: SlackInstance,
    ) -> Optional[CorpUserEditableInfoClass]:
        """
        Populate CorpUserEditableInfo aspect with user information from Slack.
        If changes are not required, None is returned.
        If changes are required, the updated aspect is returned.
        """
        mutation_required = False
        if not corpuser_editable_info.email and user_obj.email:
            mutation_required = True
            corpuser_editable_info.email = user_obj.email
        if not corpuser_editable_info.slack and user_obj.slack_id:
            mutation_required = True
            corpuser_editable_info.slack = user_obj.slack_id
        if not corpuser_editable_info.title and user_obj.title:
            mutation_required = True
            corpuser_editable_info.title = user_obj.title
        if user_obj.image_url and (
            is_picture_default_or_missing(corpuser_editable_info.pictureLink)
            or (
                is_slack_image(corpuser_editable_info.pictureLink)
                and user_obj.image_url != corpuser_editable_info.pictureLink
            )
        ):
            mutation_required = True
            corpuser_editable_info.pictureLink = user_obj.image_url
        if user_obj.phone and not corpuser_editable_info.phone:
            mutation_required = True
            corpuser_editable_info.phone = user_obj.phone
        if (
            not corpuser_editable_info.displayName
            or corpuser_editable_info.displayName == corpuser_editable_info.email
        ) and user_obj.real_name:
            mutation_required = True
            corpuser_editable_info.displayName = user_obj.real_name
        if mutation_required:
            # update informationSources
            corpuser_editable_info.informationSources = (
                []
                if not corpuser_editable_info.informationSources
                else corpuser_editable_info.informationSources
            )
            if platform_resource_urn not in corpuser_editable_info.informationSources:
                corpuser_editable_info.informationSources.append(platform_resource_urn)
            return corpuser_editable_info
        return None

    def get_public_channels(self) -> Iterable[MetadataWorkUnit]:
        cursor = None
        while True:
            with self.rate_limiter:
                channels, cursor = self._get_channel_info(cursor=cursor)
            for channel in channels:
                self.report.channels_reported += 1
                yield channel
            if not cursor:
                break

    def emit_slack_member_aspect(
        self, user: SlackUserInfo
    ) -> Iterable[MetadataWorkUnit]:
        slack_user = SlackUserDetails(slack_user_info=user)
        for mcp in slack_user.to_mcps():
            yield mcp.as_workunit()

    def emit_corp_user_slack_settings(
        self, user_obj: CorpUser
    ) -> Iterable[MetadataWorkUnit]:
        assert self.ctx.graph is not None

        if not user_obj.urn:
            return

        corp_user_settings = self.ctx.graph.get_aspect(
            user_obj.urn, CorpUserSettingsClass
        )
        if not corp_user_settings:
            return

        notification_settings = corp_user_settings.notificationSettings

        if not notification_settings:
            corp_user_settings.notificationSettings = NotificationSettingsClass(
                sinkTypes=[],
                slackSettings=SlackNotificationSettingsClass(
                    userHandle=user_obj.slack_id
                ),
            )
        elif (
            not notification_settings.slackSettings
            or not notification_settings.slackSettings.userHandle
        ):
            notification_settings.slackSettings = SlackNotificationSettingsClass(
                userHandle=user_obj.slack_id
            )
        else:
            return

        yield MetadataWorkUnit(
            id=f"{user_obj.urn}",
            mcp=MetadataChangeProposalWrapper(
                entityUrn=user_obj.urn,
                aspect=corp_user_settings,
            ),
        )

    @retry(
        wait=wait_exponential(multiplier=2, min=4, max=60),
        before_sleep=before_sleep_log(logger, logging.ERROR, True),
    )
    def get_user_to_be_updated(
        self,
    ) -> Iterable[Tuple[CorpUser, Optional[CorpUserEditableInfoClass]]]:
        assert self.ctx.graph is not None
        for urn in self.ctx.graph.get_urns_by_filter(
            entity_types=["corpuser"], query="*"
        ):
            user_obj = CorpUser()
            user_obj.urn = urn
            editable_properties = self.ctx.graph.get_aspect(
                urn, CorpUserEditableInfoClass
            )
            if editable_properties and editable_properties.email:
                user_obj.email = editable_properties.email
            else:
                urn_id = Urn.from_string(user_obj.urn).get_entity_id_as_string()
                if "@" in urn_id:
                    user_obj.email = urn_id
            if user_obj.email is not None:
                yield (user_obj, editable_properties)

    def get_report(self) -> SourceReport:
        return self.report



import logging
from contextlib import AbstractContextManager
from dataclasses import dataclass, field
from datetime import datetime, timezone

from datahub.utilities.perf_timer import PerfTimer
from datahub.utilities.stats_collections import TopKDict

logger: logging.Logger = logging.getLogger(__name__)


METADATA_EXTRACTION = "Metadata Extraction"
LINEAGE_EXTRACTION = "Lineage Extraction"
USAGE_EXTRACTION_INGESTION = "Usage Extraction Ingestion"
USAGE_EXTRACTION_OPERATIONAL_STATS = "Usage Extraction Operational Stats"
USAGE_EXTRACTION_USAGE_AGGREGATION = "Usage Extraction Usage Aggregation"
EXTERNAL_TABLE_DDL_LINEAGE = "External table DDL Lineage"
VIEW_PARSING = "View Parsing"
QUERIES_EXTRACTION = "Queries Extraction"
PROFILING = "Profiling"


@dataclass
class IngestionStageReport:
    ingestion_stage_durations: TopKDict[str, float] = field(default_factory=TopKDict)

    def new_stage(self, stage: str) -> "IngestionStageContext":
        return IngestionStageContext(stage, self)


@dataclass
class IngestionStageContext(AbstractContextManager):
    def __init__(self, stage: str, report: IngestionStageReport):
        self._ingestion_stage = f"{stage} at {datetime.now(timezone.utc)}"
        self._timer: PerfTimer = PerfTimer()
        self._report = report

    def __enter__(self) -> "IngestionStageContext":
        logger.info(f"Stage started: {self._ingestion_stage}")
        self._timer.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        elapsed = self._timer.elapsed_seconds(digits=2)
        logger.info(
            f"Time spent in stage <{self._ingestion_stage}>: {elapsed} seconds",
            stacklevel=2,
        )
        self._report.ingestion_stage_durations[self._ingestion_stage] = elapsed
        return None
